"""

    NOTE: This is a code generated file.  This file should not be edited directly.
"""



from akit.aspects import AspectsUPnP, DEFAULT_UPNP_ASPECTS

from akit.extensible import LoadableExtension
from akit.interop.upnp.services.upnpserviceproxy import UpnpServiceProxy

class WLANConfiguration1ServiceProxy(UpnpServiceProxy, LoadableExtension):
    """
        This is a code generated proxy class to the 'urn:schemas-upnp-org:service:WLANConfiguration:1' service.
    """

    SERVICE_MANUFACTURER = 'UPnP'
    SERVICE_TYPE = 'WLANConfiguration1'

    SERVICE_DEFAULT_VARIABLES = {
        "AssociatedDeviceAuthenticationState": { "data_type": "boolean", "default": None, "allowed_list": None},
        "AssociatedDeviceIPAddress": { "data_type": "string", "default": None, "allowed_list": None},
        "AssociatedDeviceMACAddress": { "data_type": "string", "default": None, "allowed_list": None},
        "AuthenticationServiceMode": { "data_type": "string", "default": None, "allowed_list": "['None', 'LinkAuthentication', 'RadiusClient']"},
        "AutoRateFallBackEnabled": { "data_type": "boolean", "default": None, "allowed_list": None},
        "BSSID": { "data_type": "string", "default": None, "allowed_list": None},
        "BasicAuthenticationMode": { "data_type": "string", "default": None, "allowed_list": "['None', 'EAPAuthentication']"},
        "BasicDataTransmissionRates": { "data_type": "string", "default": None, "allowed_list": None},
        "BasicEncryptionModes": { "data_type": "string", "default": None, "allowed_list": "['None', 'WEPEncryption']"},
        "BeaconAdvertisementEnabled": { "data_type": "boolean", "default": None, "allowed_list": None},
        "BeaconType": { "data_type": "string", "default": None, "allowed_list": "['None', 'Basic', 'WPA', '11i', 'BasicandWPA', 'Basicand11i', 'WPAand11i', 'BasicandWPAand11i']"},
        "Channel": { "data_type": "ui1", "default": None, "allowed_list": None},
        "ChannelsInUse": { "data_type": "string", "default": None, "allowed_list": None},
        "DeviceOperationMode": { "data_type": "string", "default": None, "allowed_list": "['InfrastructureAccessPoint', 'WirelessBridgePointToPoint', 'WirelessBridgePointToMultipoint', 'WirelessRepeater', 'WirelessSTA']"},
        "DistanceFromRoot": { "data_type": "ui1", "default": None, "allowed_list": None},
        "IEEE11iAuthenticationMode": { "data_type": "string", "default": None, "allowed_list": "['TKIPAuthentication', 'EAPAuthentication', 'EAPandTKIPAuthentication']"},
        "IEEE11iEncryptionModes": { "data_type": "string", "default": None, "allowed_list": "['WEPEncryption', 'TKIPEncryption', 'WEPandTKIPEncryption', 'AESEncryption', 'WEPandAESEncryption', 'TKIPandAESEncryption', 'WEPandTKIPandAESEncryption']"},
        "InsecureOutOfBandAccessEnabled": { "data_type": "boolean", "default": None, "allowed_list": None},
        "KeyPassphrase": { "data_type": "string", "default": None, "allowed_list": None},
        "LocationDescription": { "data_type": "string", "default": None, "allowed_list": None},
        "OperationalDataTransmissionRates": { "data_type": "string", "default": None, "allowed_list": None},
        "PeerBSSID": { "data_type": "string", "default": None, "allowed_list": None},
        "PossibleChannels": { "data_type": "string", "default": None, "allowed_list": None},
        "PossibleDataTransmissionRates": { "data_type": "string", "default": None, "allowed_list": None},
        "PreSharedKey": { "data_type": "string", "default": None, "allowed_list": None},
        "PreSharedKeyIndex": { "data_type": "ui1", "default": None, "allowed_list": None},
        "RadioEnabled": { "data_type": "boolean", "default": None, "allowed_list": None},
        "RegulatoryDomain": { "data_type": "string", "default": None, "allowed_list": None},
        "SSID": { "data_type": "string", "default": None, "allowed_list": None},
        "TotalBytesReceived": { "data_type": "ui4", "default": None, "allowed_list": None},
        "TotalBytesSent": { "data_type": "ui4", "default": None, "allowed_list": None},
        "TotalIntegrityFailures": { "data_type": "ui4", "default": None, "allowed_list": None},
        "TotalPSKFailures": { "data_type": "ui4", "default": None, "allowed_list": None},
        "TotalPacketsReceived": { "data_type": "ui4", "default": None, "allowed_list": None},
        "TotalPacketsSent": { "data_type": "ui4", "default": None, "allowed_list": None},
        "WEPEncryptionLevel": { "data_type": "string", "default": None, "allowed_list": "['Disabled', '40-bit', '104-bit']"},
        "WEPKey": { "data_type": "string", "default": None, "allowed_list": None},
        "WEPKeyIndex": { "data_type": "ui1", "default": None, "allowed_list": None},
        "WPAAuthenticationMode": { "data_type": "string", "default": None, "allowed_list": "['PSKAuthentication', 'EAPAuthentication']"},
        "WPAEncryptionModes": { "data_type": "string", "default": None, "allowed_list": "['WEPEncryption', 'TKIPEncryption', 'WEPandTKIPEncryption', 'AESEncryption', 'WEPandAESEncryption', 'TKIPandAESEncryption', 'WEPandTKIPandAESEncryption']"},
    }

    SERVICE_EVENT_VARIABLES = {
        "TotalAssociations": { "data_type": "ui2", "default": None, "allowed_list": None},
    }

    def action_FactoryDefaultReset(self, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the FactoryDefaultReset action.
        """
        arguments = { }

        self.call_action("FactoryDefaultReset", arguments=arguments, aspects=aspects)

        return

    def action_Get11iBeaconSecurityProperties(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the Get11iBeaconSecurityProperties action.

            :returns: "NewIEEE11iEncryptionModes", "NewIEEE11iAuthenticationMode"
        """
        arguments = { }

        out_params = self.call_action("Get11iBeaconSecurityProperties", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewIEEE11iEncryptionModes", "NewIEEE11iAuthenticationMode",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetAuthenticationServiceMode(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetAuthenticationServiceMode action.

            :returns: "NewAuthenticationServiceMode"
        """
        arguments = { }

        out_params = self.call_action("GetAuthenticationServiceMode", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewAuthenticationServiceMode",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetAutoRateFallBackMode(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetAutoRateFallBackMode action.

            :returns: "NewAutoRateFallBackEnabled"
        """
        arguments = { }

        out_params = self.call_action("GetAutoRateFallBackMode", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewAutoRateFallBackEnabled",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetBSSID(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetBSSID action.

            :returns: "NewBSSID"
        """
        arguments = { }

        out_params = self.call_action("GetBSSID", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewBSSID",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetBasicBeaconSecurityProperties(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetBasicBeaconSecurityProperties action.

            :returns: "NewBasicEncryptionModes", "NewBasicAuthenticationMode"
        """
        arguments = { }

        out_params = self.call_action("GetBasicBeaconSecurityProperties", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewBasicEncryptionModes", "NewBasicAuthenticationMode",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetBeaconAdvertisement(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetBeaconAdvertisement action.

            :returns: "NewBeaconAdvertisementEnabled"
        """
        arguments = { }

        out_params = self.call_action("GetBeaconAdvertisement", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewBeaconAdvertisementEnabled",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetBeaconType(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetBeaconType action.

            :returns: "NewBeaconType"
        """
        arguments = { }

        out_params = self.call_action("GetBeaconType", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewBeaconType",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetByteStatistics(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetByteStatistics action.

            :returns: "NewTotalBytesSent", "NewTotalBytesReceived"
        """
        arguments = { }

        out_params = self.call_action("GetByteStatistics", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewTotalBytesSent", "NewTotalBytesReceived",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetChannelInfo(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetChannelInfo action.

            :returns: "NewChannel", "NewPossibleChannels"
        """
        arguments = { }

        out_params = self.call_action("GetChannelInfo", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewChannel", "NewPossibleChannels",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetChannelsInUse(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetChannelsInUse action.

            :returns: "NewChannelsInUse"
        """
        arguments = { }

        out_params = self.call_action("GetChannelsInUse", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewChannelsInUse",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetDataTransmissionRateInfo(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetDataTransmissionRateInfo action.

            :returns: "NewBasicDataTransmissionRates", "NewOperationalDataTransmissionRates", "NewPossibleDataTransmissionRates"
        """
        arguments = { }

        out_params = self.call_action("GetDataTransmissionRateInfo", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewBasicDataTransmissionRates", "NewOperationalDataTransmissionRates", "NewPossibleDataTransmissionRates",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetDefaultWEPKeyIndex(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetDefaultWEPKeyIndex action.

            :returns: "NewDefaultWEPKeyIndex"
        """
        arguments = { }

        out_params = self.call_action("GetDefaultWEPKeyIndex", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewDefaultWEPKeyIndex",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetDeviceOperationMode(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetDeviceOperationMode action.

            :returns: "NewDeviceOperationMode", "NewSSID", "NewBSSID", "NewChannel", "NewBasicDataTransmissionRates", "NewOperationalDataTransmissionRates", "NewDistanceFromRoot"
        """
        arguments = { }

        out_params = self.call_action("GetDeviceOperationMode", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewDeviceOperationMode", "NewSSID", "NewBSSID", "NewChannel", "NewBasicDataTransmissionRates", "NewOperationalDataTransmissionRates", "NewDistanceFromRoot",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetFailureStatusInfo(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetFailureStatusInfo action.

            :returns: "NewTotalIntegrityFailures", "NewTotalPSKFailures"
        """
        arguments = { }

        out_params = self.call_action("GetFailureStatusInfo", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewTotalIntegrityFailures", "NewTotalPSKFailures",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetGenericAssociatedDeviceInfo(self, NewAssociatedDeviceIndex, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetGenericAssociatedDeviceInfo action.

            :returns: "NewAssociatedDeviceMACAddress", "NewAssociatedDeviceIPAddress", "NewAssociatedDeviceAuthenticationState"
        """
        arguments = {
            "NewAssociatedDeviceIndex": NewAssociatedDeviceIndex,
        }

        out_params = self.call_action("GetGenericAssociatedDeviceInfo", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewAssociatedDeviceMACAddress", "NewAssociatedDeviceIPAddress", "NewAssociatedDeviceAuthenticationState",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetInsecureOutOfBandAccessMode(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetInsecureOutOfBandAccessMode action.

            :returns: "NewInsecureOutOfBandAccessEnabled"
        """
        arguments = { }

        out_params = self.call_action("GetInsecureOutOfBandAccessMode", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewInsecureOutOfBandAccessEnabled",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetLocationDescription(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetLocationDescription action.

            :returns: "NewLocationDescription"
        """
        arguments = { }

        out_params = self.call_action("GetLocationDescription", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewLocationDescription",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetPacketStatistics(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetPacketStatistics action.

            :returns: "NewTotalPacketsSent", "NewTotalPacketsReceived"
        """
        arguments = { }

        out_params = self.call_action("GetPacketStatistics", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewTotalPacketsSent", "NewTotalPacketsReceived",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetPreSharedKey(self, NewPreSharedKeyIndex, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetPreSharedKey action.

            :returns: "NewPreSharedKey", "NewPSKPassphrase"
        """
        arguments = {
            "NewPreSharedKeyIndex": NewPreSharedKeyIndex,
        }

        out_params = self.call_action("GetPreSharedKey", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewPreSharedKey", "NewPSKPassphrase",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetRadioMode(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetRadioMode action.

            :returns: "NewRadioEnabled"
        """
        arguments = { }

        out_params = self.call_action("GetRadioMode", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewRadioEnabled",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetRegulatoryDomain(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetRegulatoryDomain action.

            :returns: "NewRegulatoryDomain"
        """
        arguments = { }

        out_params = self.call_action("GetRegulatoryDomain", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewRegulatoryDomain",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetSSID(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetSSID action.

            :returns: "NewSSID"
        """
        arguments = { }

        out_params = self.call_action("GetSSID", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewSSID",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetSecurityKeys(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetSecurityKeys action.

            :returns: "NewWEPKey0", "NewWEPKey1", "NewWEPKey2", "NewWEPKey3", "NewPreSharedKey", "NewKeyPassphrase"
        """
        arguments = { }

        out_params = self.call_action("GetSecurityKeys", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewWEPKey0", "NewWEPKey1", "NewWEPKey2", "NewWEPKey3", "NewPreSharedKey", "NewKeyPassphrase",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetSpecificAssociatedDeviceInfo(self, NewAssociatedDeviceMACAddress, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetSpecificAssociatedDeviceInfo action.

            :returns: "NewAssociatedDeviceIPAddress", "NewAssociatedDeviceAuthenticationState"
        """
        arguments = {
            "NewAssociatedDeviceMACAddress": NewAssociatedDeviceMACAddress,
        }

        out_params = self.call_action("GetSpecificAssociatedDeviceInfo", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewAssociatedDeviceIPAddress", "NewAssociatedDeviceAuthenticationState",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetTotalAssociations(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetTotalAssociations action.

            :returns: "NewTotalAssociations"
        """
        arguments = { }

        out_params = self.call_action("GetTotalAssociations", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewTotalAssociations",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetWPABeaconSecurityProperties(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetWPABeaconSecurityProperties action.

            :returns: "NewWPAEncryptionModes", "NewWPAAuthenticationMode"
        """
        arguments = { }

        out_params = self.call_action("GetWPABeaconSecurityProperties", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewWPAEncryptionModes", "NewWPAAuthenticationMode",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_ResetAuthentication(self, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the ResetAuthentication action.
        """
        arguments = { }

        self.call_action("ResetAuthentication", arguments=arguments, aspects=aspects)

        return

    def action_Set11iBeaconSecurityProperties(self, NewIEEE11iEncryptionModes, NewIEEE11iAuthenticationMode, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the Set11iBeaconSecurityProperties action.
        """
        arguments = {
            "NewIEEE11iEncryptionModes": NewIEEE11iEncryptionModes,
            "NewIEEE11iAuthenticationMode": NewIEEE11iAuthenticationMode,
        }

        self.call_action("Set11iBeaconSecurityProperties", arguments=arguments, aspects=aspects)

        return

    def action_SetAuthenticationServiceMode(self, NewAuthenticationServiceMode, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetAuthenticationServiceMode action.
        """
        arguments = {
            "NewAuthenticationServiceMode": NewAuthenticationServiceMode,
        }

        self.call_action("SetAuthenticationServiceMode", arguments=arguments, aspects=aspects)

        return

    def action_SetAutoRateFallBackMode(self, NewAutoRateFallBackEnabled, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetAutoRateFallBackMode action.
        """
        arguments = {
            "NewAutoRateFallBackEnabled": NewAutoRateFallBackEnabled,
        }

        self.call_action("SetAutoRateFallBackMode", arguments=arguments, aspects=aspects)

        return

    def action_SetBasicBeaconSecurityProperties(self, NewBasicEncryptionModes, NewBasicAuthenticationMode, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetBasicBeaconSecurityProperties action.
        """
        arguments = {
            "NewBasicEncryptionModes": NewBasicEncryptionModes,
            "NewBasicAuthenticationMode": NewBasicAuthenticationMode,
        }

        self.call_action("SetBasicBeaconSecurityProperties", arguments=arguments, aspects=aspects)

        return

    def action_SetBeaconAdvertisement(self, NewBeaconAdvertisementEnabled, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetBeaconAdvertisement action.
        """
        arguments = {
            "NewBeaconAdvertisementEnabled": NewBeaconAdvertisementEnabled,
        }

        self.call_action("SetBeaconAdvertisement", arguments=arguments, aspects=aspects)

        return

    def action_SetBeaconType(self, NewBeaconType, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetBeaconType action.
        """
        arguments = {
            "NewBeaconType": NewBeaconType,
        }

        self.call_action("SetBeaconType", arguments=arguments, aspects=aspects)

        return

    def action_SetChannel(self, NewChannel, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetChannel action.
        """
        arguments = {
            "NewChannel": NewChannel,
        }

        self.call_action("SetChannel", arguments=arguments, aspects=aspects)

        return

    def action_SetDataTransmissionRates(self, NewBasicDataTransmissionRates, NewOperationalDataTransmissionRates, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetDataTransmissionRates action.
        """
        arguments = {
            "NewBasicDataTransmissionRates": NewBasicDataTransmissionRates,
            "NewOperationalDataTransmissionRates": NewOperationalDataTransmissionRates,
        }

        self.call_action("SetDataTransmissionRates", arguments=arguments, aspects=aspects)

        return

    def action_SetDefaultWEPKeyIndex(self, NewDefaultWEPKeyIndex, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetDefaultWEPKeyIndex action.
        """
        arguments = {
            "NewDefaultWEPKeyIndex": NewDefaultWEPKeyIndex,
        }

        self.call_action("SetDefaultWEPKeyIndex", arguments=arguments, aspects=aspects)

        return

    def action_SetDeviceOperationMode(self, NewDeviceOperationMode, NewSSID, NewBSSID, NewChannel, NewBasicDataTransmissionRates, NewOperationalDataTransmissionRates, NewDistanceFromRoot, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetDeviceOperationMode action.
        """
        arguments = {
            "NewDeviceOperationMode": NewDeviceOperationMode,
            "NewSSID": NewSSID,
            "NewBSSID": NewBSSID,
            "NewChannel": NewChannel,
            "NewBasicDataTransmissionRates": NewBasicDataTransmissionRates,
            "NewOperationalDataTransmissionRates": NewOperationalDataTransmissionRates,
            "NewDistanceFromRoot": NewDistanceFromRoot,
        }

        self.call_action("SetDeviceOperationMode", arguments=arguments, aspects=aspects)

        return

    def action_SetInsecureOutOfBandAccessMode(self, NewInsecureOutOfBandAccessEnabled, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetInsecureOutOfBandAccessMode action.
        """
        arguments = {
            "NewInsecureOutOfBandAccessEnabled": NewInsecureOutOfBandAccessEnabled,
        }

        self.call_action("SetInsecureOutOfBandAccessMode", arguments=arguments, aspects=aspects)

        return

    def action_SetLocationDescription(self, NewLocationDescription, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetLocationDescription action.
        """
        arguments = {
            "NewLocationDescription": NewLocationDescription,
        }

        self.call_action("SetLocationDescription", arguments=arguments, aspects=aspects)

        return

    def action_SetPreSharedKey(self, NewPreSharedKeyIndex, NewPreSharedKey, NewPSKPassphrase, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetPreSharedKey action.
        """
        arguments = {
            "NewPreSharedKeyIndex": NewPreSharedKeyIndex,
            "NewPreSharedKey": NewPreSharedKey,
            "NewPSKPassphrase": NewPSKPassphrase,
        }

        self.call_action("SetPreSharedKey", arguments=arguments, aspects=aspects)

        return

    def action_SetRadioMode(self, NewRadioEnabled, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetRadioMode action.
        """
        arguments = {
            "NewRadioEnabled": NewRadioEnabled,
        }

        self.call_action("SetRadioMode", arguments=arguments, aspects=aspects)

        return

    def action_SetRegulatoryDomain(self, NewRegulatoryDomain, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetRegulatoryDomain action.
        """
        arguments = {
            "NewRegulatoryDomain": NewRegulatoryDomain,
        }

        self.call_action("SetRegulatoryDomain", arguments=arguments, aspects=aspects)

        return

    def action_SetSSID(self, NewSSID, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetSSID action.
        """
        arguments = {
            "NewSSID": NewSSID,
        }

        self.call_action("SetSSID", arguments=arguments, aspects=aspects)

        return

    def action_SetSecurityKeys(self, NewWEPKey0, NewWEPKey1, NewWEPKey2, NewWEPKey3, NewPreSharedKey, NewKeyPassphrase, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetSecurityKeys action.
        """
        arguments = {
            "NewWEPKey0": NewWEPKey0,
            "NewWEPKey1": NewWEPKey1,
            "NewWEPKey2": NewWEPKey2,
            "NewWEPKey3": NewWEPKey3,
            "NewPreSharedKey": NewPreSharedKey,
            "NewKeyPassphrase": NewKeyPassphrase,
        }

        self.call_action("SetSecurityKeys", arguments=arguments, aspects=aspects)

        return

    def action_SetWPABeaconSecurityProperties(self, NewWPAEncryptionModes, NewWPAAuthenticationMode, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetWPABeaconSecurityProperties action.
        """
        arguments = {
            "NewWPAEncryptionModes": NewWPAEncryptionModes,
            "NewWPAAuthenticationMode": NewWPAAuthenticationMode,
        }

        self.call_action("SetWPABeaconSecurityProperties", arguments=arguments, aspects=aspects)

        return



"""

    NOTE: This is a code generated file.  This file should not be edited directly.
"""



from akit.extensible import LoadableExtension
from akit.interop.upnp.services.upnpserviceproxy import UpnpServiceProxy

class AVTransport1ServiceProxy(UpnpServiceProxy, LoadableExtension):
    """
        This is a code generated proxy class to the 'AVTransport1' service.
    """

    SERVICE_MANUFACTURER = 'MicrosoftCorporation'
    SERVICE_TYPE = 'urn:schemas-upnp-org:service:AVTransport:1'

    SERVICE_EVENT_VARIABLES = {
        "LastChange": { "data_type": "string", "default": None, "allowed_list": None},
    }

    def action_GetCurrentTransportActions(self, InstanceID, extract_returns=True):
        """
            Calls the GetCurrentTransportActions action.

            :returns: "Actions"
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        out_params = self._proxy_call_action("GetCurrentTransportActions", arguments=arguments)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("Actions",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetDeviceCapabilities(self, InstanceID, extract_returns=True):
        """
            Calls the GetDeviceCapabilities action.

            :returns: "PlayMedia", "RecMedia", "RecQualityModes"
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        out_params = self._proxy_call_action("GetDeviceCapabilities", arguments=arguments)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("PlayMedia", "RecMedia", "RecQualityModes",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetMediaInfo(self, InstanceID, extract_returns=True):
        """
            Calls the GetMediaInfo action.

            :returns: "NrTracks", "MediaDuration", "CurrentURI", "CurrentURIMetaData", "NextURI", "NextURIMetaData", "PlayMedium", "RecordMedium", "WriteStatus"
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        out_params = self._proxy_call_action("GetMediaInfo", arguments=arguments)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NrTracks", "MediaDuration", "CurrentURI", "CurrentURIMetaData", "NextURI", "NextURIMetaData", "PlayMedium", "RecordMedium", "WriteStatus",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetPositionInfo(self, InstanceID, extract_returns=True):
        """
            Calls the GetPositionInfo action.

            :returns: "Track", "TrackDuration", "TrackMetaData", "TrackURI", "RelTime", "AbsTime", "RelCount", "AbsCount"
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        out_params = self._proxy_call_action("GetPositionInfo", arguments=arguments)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("Track", "TrackDuration", "TrackMetaData", "TrackURI", "RelTime", "AbsTime", "RelCount", "AbsCount",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetTransportInfo(self, InstanceID, extract_returns=True):
        """
            Calls the GetTransportInfo action.

            :returns: "CurrentTransportState", "CurrentTransportStatus", "CurrentSpeed"
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        out_params = self._proxy_call_action("GetTransportInfo", arguments=arguments)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentTransportState", "CurrentTransportStatus", "CurrentSpeed",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetTransportSettings(self, InstanceID, extract_returns=True):
        """
            Calls the GetTransportSettings action.

            :returns: "PlayMode", "RecQualityMode"
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        out_params = self._proxy_call_action("GetTransportSettings", arguments=arguments)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("PlayMode", "RecQualityMode",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_Next(self, InstanceID, extract_returns=True):
        """
            Calls the Next action.

            :returns: "result"
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        out_params = self._proxy_call_action("Next", arguments=arguments)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("result",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_Pause(self, InstanceID, extract_returns=True):
        """
            Calls the Pause action.

            :returns: "result"
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        out_params = self._proxy_call_action("Pause", arguments=arguments)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("result",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_Play(self, InstanceID, Speed, extract_returns=True):
        """
            Calls the Play action.

            :returns: "result"
        """
        arguments = {
            "InstanceID": InstanceID,
            "Speed": Speed,
        }

        out_params = self._proxy_call_action("Play", arguments=arguments)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("result",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_Previous(self, InstanceID, extract_returns=True):
        """
            Calls the Previous action.

            :returns: "result"
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        out_params = self._proxy_call_action("Previous", arguments=arguments)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("result",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_Seek(self, InstanceID, Unit, Target, extract_returns=True):
        """
            Calls the Seek action.

            :returns: "result"
        """
        arguments = {
            "InstanceID": InstanceID,
            "Unit": Unit,
            "Target": Target,
        }

        out_params = self._proxy_call_action("Seek", arguments=arguments)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("result",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_SetAVTransportURI(self, InstanceID, CurrentURI, CurrentURIMetaData, extract_returns=True):
        """
            Calls the SetAVTransportURI action.

            :returns: "result"
        """
        arguments = {
            "InstanceID": InstanceID,
            "CurrentURI": CurrentURI,
            "CurrentURIMetaData": CurrentURIMetaData,
        }

        out_params = self._proxy_call_action("SetAVTransportURI", arguments=arguments)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("result",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_SetNextAVTransportURI(self, InstanceID, NextURI, NextURIMetaData, extract_returns=True):
        """
            Calls the SetNextAVTransportURI action.

            :returns: "result"
        """
        arguments = {
            "InstanceID": InstanceID,
            "NextURI": NextURI,
            "NextURIMetaData": NextURIMetaData,
        }

        out_params = self._proxy_call_action("SetNextAVTransportURI", arguments=arguments)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("result",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_Stop(self, InstanceID, extract_returns=True):
        """
            Calls the Stop action.

            :returns: "result"
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        out_params = self._proxy_call_action("Stop", arguments=arguments)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("result",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args



"""

    NOTE: This is a code generated file.  This file should not be edited directly.
"""



from akit.aspects import AspectsUPnP, DEFAULT_UPNP_ASPECTS

from akit.extensible import LoadableExtension
from akit.interop.upnp.services.upnpserviceproxy import UpnpServiceProxy

class InputConfig1ServiceProxy(UpnpServiceProxy, LoadableExtension):
    """
        This is a code generated proxy class to the 'urn:schemas-upnp-org:service:InputConfig:1' service.
    """

    SERVICE_MANUFACTURER = 'UPnP'
    SERVICE_TYPE = 'InputConfig1'

    SERVICE_DEFAULT_VARIABLES = {
        "InputConnectionList": { "data_type": "string", "default": None, "allowed_list": None},
    }

    SERVICE_EVENT_VARIABLES = {
        "DeviceInputCapability": { "data_type": "string", "default": None, "allowed_list": None},
        "RequiredInputType": { "data_type": "string", "default": None, "allowed_list": None},
    }

    def action_GetInputCapability(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetInputCapability action.

            :returns: "SupportedCapabilities"
        """
        arguments = { }

        out_params = self.call_action("GetInputCapability", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("SupportedCapabilities",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetInputConnectionList(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetInputConnectionList action.

            :returns: "CurrentConnectionList"
        """
        arguments = { }

        out_params = self.call_action("GetInputConnectionList", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentConnectionList",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_SetInputSession(self, SelectedCapability, ReceiverInfo, PeerDeviceInfo, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetInputSession action.

            :returns: "SessionID", "ConnectionInfo"
        """
        arguments = {
            "SelectedCapability": SelectedCapability,
            "ReceiverInfo": ReceiverInfo,
            "PeerDeviceInfo": PeerDeviceInfo,
        }

        out_params = self.call_action("SetInputSession", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("SessionID", "ConnectionInfo",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_SetMonopolizedSender(self, OwnerDeviceInfo, OwnedSessionID, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetMonopolizedSender action.
        """
        arguments = {
            "OwnerDeviceInfo": OwnerDeviceInfo,
            "OwnedSessionID": OwnedSessionID,
        }

        self.call_action("SetMonopolizedSender", arguments=arguments, aspects=aspects)

        return

    def action_SetMultiInputMode(self, NewMultiInputMode, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetMultiInputMode action.
        """
        arguments = {
            "NewMultiInputMode": NewMultiInputMode,
        }

        self.call_action("SetMultiInputMode", arguments=arguments, aspects=aspects)

        return

    def action_StartInputSession(self, SessionID, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the StartInputSession action.
        """
        arguments = {
            "SessionID": SessionID,
        }

        self.call_action("StartInputSession", arguments=arguments, aspects=aspects)

        return

    def action_StopInputsession(self, SessionID, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the StopInputsession action.
        """
        arguments = {
            "SessionID": SessionID,
        }

        self.call_action("StopInputsession", arguments=arguments, aspects=aspects)

        return

    def action_SwitchInputSession(self, SessionID, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SwitchInputSession action.
        """
        arguments = {
            "SessionID": SessionID,
        }

        self.call_action("SwitchInputSession", arguments=arguments, aspects=aspects)

        return




"""
.. module:: akit.interop.upnp.extensions.sonos.zps23
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Module containing the a Upnp device for a Sonos Zps23.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>

"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from akit.interop.upnp.devices.upnpdevice import UpnpDevice
from akit.interop.upnp.devices.upnprootdevice import UpnpRootDevice
from akit.interop.upnp.xml.upnpdevice1 import UpnpDevice1Device


from akit.interop.upnp.extensions.standard.rootdevices.sonos.sonosdevice import SonosDevice, SonosUpnpDevice1Device

from akit.interop.upnp.extensions.dynamic.services.SonosInc.alarmclock1serviceproxy import AlarmClock1ServiceProxy
from akit.interop.upnp.extensions.dynamic.services.SonosInc.audioin1serviceproxy import AudioIn1ServiceProxy
from akit.interop.upnp.extensions.dynamic.services.SonosInc.avtransport1serviceproxy import AVTransport1ServiceProxy
from akit.interop.upnp.extensions.dynamic.services.SonosInc.connectionmanager1serviceproxy import ConnectionManager1ServiceProxy
from akit.interop.upnp.extensions.dynamic.services.SonosInc.contentdirectory1serviceproxy import ContentDirectory1ServiceProxy
from akit.interop.upnp.extensions.dynamic.services.SonosInc.deviceproperties1serviceproxy import DeviceProperties1ServiceProxy
from akit.interop.upnp.extensions.dynamic.services.SonosInc.groupmanagement1serviceproxy import GroupManagement1ServiceProxy
from akit.interop.upnp.extensions.dynamic.services.SonosInc.grouprenderingcontrol1serviceproxy import GroupRenderingControl1ServiceProxy
from akit.interop.upnp.extensions.dynamic.services.SonosInc.htcontrol1serviceproxy import HTControl1ServiceProxy
from akit.interop.upnp.extensions.dynamic.services.SonosInc.musicservices1serviceproxy import MusicServices1ServiceProxy
from akit.interop.upnp.extensions.dynamic.services.SonosInc.qplay1serviceproxy import QPlay1ServiceProxy
from akit.interop.upnp.extensions.dynamic.services.SonosInc.queue1serviceproxy import Queue1ServiceProxy
from akit.interop.upnp.extensions.dynamic.services.SonosInc.renderingcontrol1serviceproxy import RenderingControl1ServiceProxy
from akit.interop.upnp.extensions.dynamic.services.SonosInc.systemproperties1serviceproxy import SystemProperties1ServiceProxy
from akit.interop.upnp.extensions.dynamic.services.SonosInc.virtuallinein1serviceproxy import VirtualLineIn1ServiceProxy
from akit.interop.upnp.extensions.dynamic.services.SonosInc.zonegrouptopology1serviceproxy import ZoneGroupTopology1ServiceProxy

class SonosPlayer(SonosDevice):
    """
    """

    SERVICE_NAMES = [
        AlarmClock1ServiceProxy.SERVICE_TYPE,
        AudioIn1ServiceProxy.SERVICE_TYPE,
        AVTransport1ServiceProxy.SERVICE_TYPE,
        ConnectionManager1ServiceProxy.SERVICE_TYPE,
        ContentDirectory1ServiceProxy.SERVICE_TYPE,
        DeviceProperties1ServiceProxy.SERVICE_TYPE,
        GroupManagement1ServiceProxy.SERVICE_TYPE,
        GroupRenderingControl1ServiceProxy.SERVICE_TYPE,
        HTControl1ServiceProxy.SERVICE_TYPE,
        MusicServices1ServiceProxy.SERVICE_TYPE,
        QPlay1ServiceProxy.SERVICE_TYPE,
        Queue1ServiceProxy.SERVICE_TYPE,
        RenderingControl1ServiceProxy.SERVICE_TYPE,
        SystemProperties1ServiceProxy.SERVICE_TYPE,
        VirtualLineIn1ServiceProxy.SERVICE_TYPE,
        ZoneGroupTopology1ServiceProxy.SERVICE_TYPE
    ]

    def __init__(self, manufacturer: str, modelNumber: str, modelDescription: str):
        super(SonosPlayer, self).__init__(manufacturer, modelNumber, modelDescription)

        self._bood_id = None
        self._boot_seq = None
        self._household = None
        self._smart_speaker = None
        self._variant = None
        self._wifimode = None
        return

    @property
    def bootId(self):
        return self._bood_id

    @property
    def bootSeq(self):
        return self._boot_seq

    @property
    def household(self):
        return self._household

    @property
    def smartSpeaker(self):
        return self._smart_speaker

    @property
    def variant(self):
        return self._variant

    @property
    def wifiMode(self):
        return self._wifimode

    def deviceMediaRenderer(self):
        return

    def deviceMediaServer(self):
        return

    def getLedState(self):
        devprops = self.serviceDeviceProperties()
        rtnval = devprops.action_GetLEDState()
        return rtnval

    def serviceAlarmClock(self, allow_none: bool=False):
        svctype = 'urn:schemas-upnp-org:service:AlarmClock:1'
        svc = self.lookup_service(self.MANUFACTURER, svctype, allow_none=allow_none)
        return svc

    def serviceAVTransport(self, allow_none: bool=False):
        svctype = 'urn:schemas-upnp-org:service:AVTransport:1'
        svc = self.lookup_service(self.MANUFACTURER, svctype, allow_none=allow_none)
        return svc

    def serviceDeviceProperties(self, allow_none: bool=False):
        svctype = 'urn:schemas-upnp-org:service:DeviceProperties:1'
        svc = self.lookup_service(self.MANUFACTURER, svctype, allow_none=allow_none)
        return svc

    def serviceGroupManagement(self, allow_none: bool=False):
        svctype = 'urn:schemas-upnp-org:service:GroupManagement:1'
        svc = self.lookup_service(self.MANUFACTURER, svctype, allow_none=allow_none)
        return svc

    def serviceMusicService(self, allow_none: bool=False):
        svctype = 'urn:schemas-upnp-org:service:MusicService:1'
        svc = self.lookup_service(self.MANUFACTURER, svctype, allow_none=allow_none)
        return svc

    def serviceSystemProperties(self, allow_none: bool=False):
        svctype = 'urn:schemas-upnp-org:service:SystemProperties:1'
        svc = self.lookup_service(self.MANUFACTURER, svctype, allow_none=allow_none)
        return svc

    def serviceQPlay(self, allow_none: bool=False):
        svctype = 'urn:schemas-upnp-org:service:QPlay:1'
        svc = self.lookup_service(self.MANUFACTURER, svctype, allow_none=allow_none)
        return svc

    def serviceZoneTopologyGroup(self, allow_none: bool=False):
        svctype = 'urn:schemas-upnp-org:service:ZoneTopologyGroup:1'
        svc = self.lookup_service(self.MANUFACTURER, svctype, allow_none=allow_none)
        return svc

    def _consume_upnp_extra(self, extrainfo):

        if "BOOTID.UPNP.ORG" in extrainfo:
            self._bood_id = extrainfo.pop("BOOTID.UPNP.ORG")
        if "X-RINCON-BOOTSEQ" in extrainfo:
            self._boot_seq = extrainfo.pop("X-RINCON-BOOTSEQ")
        if "X-RINCON-HOUSEHOLD" in extrainfo:
            self._household = extrainfo.pop("X-RINCON-HOUSEHOLD")
        if "HOUSEHOLD.SMARTSPEAKER.AUDIO" in extrainfo:
            self._smart_speaker = extrainfo.pop("HOUSEHOLD.SMARTSPEAKER.AUDIO")
        if "X-RINCON-VARIANT" in extrainfo:
            self._variant = extrainfo.pop("X-RINCON-VARIANT")
        if "X-RINCON-WIFIMODE" in extrainfo:
            self._wifimode = extrainfo.pop("X-RINCON-WIFIMODE")

        self._extra = extrainfo
        return

    def _create_device_description_node(self, devNode, namespaces=None):
        dev_desc_node = SonosUpnpDevice1Device(devNode, namespaces=namespaces)
        return dev_desc_node

    def _process_embedded_device_node(self, devNode, namespaces=None):
        dev = UpnpDevice1Device(devNode, namespaces=namespaces)
        return dev

    def _process_other_node(self, otherNode, namespaces=None):
        return

    def _process_servicelist_node(self, listNode, namespaces=None):
        super(SonosPlayer, self)._process_servicelist_node(listNode, namespaces=namespaces)
        return

    def to_dict(self, brief=False):
        dval = super(SonosPlayer, self).to_dict(brief=brief)

        dval["bootId"] = self.bootId
        dval["bootSeq"] = self.bootSeq
        dval["household"] = self.household
        dval["smartSpeaker"] = self.smartSpeaker
        dval["variant"] = self.variant
        dval["wifiMode"] = self.wifiMode

        return dval



"""
.. module:: powercoordinator
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Contains the PowerCoordinator which is used for managing power activity services.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"


from typing import Union, TYPE_CHECKING

from akit.exceptions import AKitConfigurationError
from akit.interop.coordinators.coordinatorbase import CoordinatorBase

from akit.interop.agents.poweragents import DliPowerAgent

import dlipower

if TYPE_CHECKING:
    from akit.interop.landscaping.landscape import Landscape

class PowerCoordinator(CoordinatorBase):

    def __init__(self, lscape: "Landscape", *args, **kwargs):
        super(PowerCoordinator, self).__init__(lscape, *args, **kwargs)
        return

    def _initialize(self, *_args, **_kwargs):
        """
            Called by the CoordinatorBase constructor to perform the one time initialization of the coordinator Singleton
            of a given type.
        """
        self._power_config = {}
        for pcfg in self._coord_config:
            cfgname = pcfg["name"]
            self._power_config[cfgname] = pcfg

        self._power_interfaces = {}
        return

    def lookup_agent(self, power_mapping: dict) -> Union[DliPowerAgent, None]:
        """
            Looks up a power agent by power mapping.
        """
        power_agent = None

        pname = power_mapping["name"]
        pswitch = power_mapping["switch"]

        power_iface = self._lookup_power_interface(pname)
        if power_iface is not None:
            power_agent = DliPowerAgent(power_iface, pswitch)
        else:
            errmsg = "Failure to find power interface %r." % pname
            raise AKitConfigurationError(errmsg) from None

        return power_agent

    def _lookup_power_interface(self, interface_name: str) -> Union[dict, None]:
        """
            Looks up a power interface by power interface name.
        """
        power_iface = None

        if interface_name in self._power_interfaces:
            power_iface = self._power_interfaces[interface_name]
        else:
            lscape = self.landscape
            interface_config = self._power_config[interface_name]
            
            powerType = interface_config["powerType"]

            if powerType == "DliPowerSwitch":
                model = interface_config["model"]
                host = interface_config["host"]

                credential_name = interface_config["credential"]
                credobj = lscape.lookup_credential(credential_name)

                power_iface = dlipower.PowerSwitch(userid=credobj.username, password=credobj.password, hostname=host)

                self._power_interfaces[interface_name] = power_iface
            else:
                errmsg = "Un-Support power interface type={}.".format(powerType)
                raise AKitConfigurationError(errmsg) from None

        return power_iface


"""

    NOTE: This is a code generated file.  This file should not be edited directly.
"""



from akit.extensible import LoadableExtension
from akit.interop.upnp.services.upnpserviceproxy import UpnpServiceProxy

class Layer3Forwarding1ServiceProxy(UpnpServiceProxy, LoadableExtension):
    """
        This is a code generated proxy class to the 'Layer3Forwarding1' service.
    """

    SERVICE_MANUFACTURER = 'NETGEAR'
    SERVICE_TYPE = 'urn:schemas-upnp-org:service:Layer3Forwarding:1'

    SERVICE_EVENT_VARIABLES = {
        "DefaultConnectionService": { "data_type": "string", "default": None, "allowed_list": None},
    }

    def action_GetDefaultConnectionService(self, extract_returns=True):
        """
            Calls the GetDefaultConnectionService action.

            :returns: "NewDefaultConnectionService"
        """
        arguments = { }

        out_params = self._proxy_call_action("GetDefaultConnectionService", arguments=arguments)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewDefaultConnectionService",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_SetDefaultConnectionService(self, NewDefaultConnectionService, extract_returns=True):
        """
            Calls the SetDefaultConnectionService action.

            :returns: "result"
        """
        arguments = {
            "NewDefaultConnectionService": NewDefaultConnectionService,
        }

        out_params = self._proxy_call_action("SetDefaultConnectionService", arguments=arguments)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("result",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args



__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

import os

import yaml

from akit.exceptions import AKitConfigurationError

from akit.paths import get_filename_for_credentials

from akit.xlogging.foundations import getAutomatonKitLogger

from akit.interop.credentials.basiccredential import BasicCredential
from akit.interop.credentials.sshcredential import SshCredential
from akit.interop.credentials.wifichoicecredential import WifiChoiceCredential

logger = getAutomatonKitLogger()

class CredentialManager:

    _instance = None

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super(CredentialManager, cls).__new__(cls, *args, **kwargs)
        return cls._instance

    def __init__(self):
        self._credentials = {}

        self._initialize_credentials()
        return

    @property
    def credentials(self):
        return self._credentials

    def _initialize_credentials(self):
        """
        """

        credential_file = get_filename_for_credentials()
        if os.path.exists(credential_file):
            credential_info = None

            with open(credential_file, 'r') as lf:
                lfcontent = lf.read()
                credential_info = yaml.safe_load(lfcontent)

            try:
                credentials_list = credential_info["credentials"]
                errors, warnings = self._validate_credentials(credentials_list)

                if len(errors) == 0:
                    for credential in credentials_list:
                        if "identifier" not in credential:
                            raise AKitConfigurationError("Credential items in 'environment/credentials' must have an 'identifier' member.") from None
                        ident = credential["identifier"]

                        if "category" not in credential:
                            raise AKitConfigurationError("Credential items in 'environment/credentials' must have an 'category' member.") from None
                        category = credential["category"]

                        if category == "basic":
                            BasicCredential.validate(credential)
                            credobj = BasicCredential(**credential)
                            self._credentials[ident] = credobj
                        elif category == "ssh":
                            SshCredential.validate(credential)
                            credobj = SshCredential(**credential)
                            self._credentials[ident] = credobj
                        elif category == "wifi-choice":
                            WifiChoiceCredential.validate(credential)
                            credobj = WifiChoiceCredential(**credential)
                            self._credentials[ident] = credobj
                        else:
                            warnmsg = "Unknown category '{}' found in credential '{}'".format(category, ident)
                            logger.warn(warnmsg)
                else:
                    errmsg_lines = [
                        "Errors found in credential file={}".format(credential_file),
                        "ERRORS:"
                    ]
                    for err in errors:
                        errmsg_lines.append("    {}".format(err))

                    errmsg_lines.append("WARNINGS:")
                    for warn in warnings:
                        errmsg_lines.append("    {}".format(warn))

                    errmsg = os.linesep.join(errmsg_lines)
                    raise AKitConfigurationError(errmsg)
            except KeyError:
                errmsg = "No 'credentials' field found in file={}".format(credential_file)
                raise AKitConfigurationError(errmsg)
        else:
            warnmsg = "Credential file not found. expected={}".format(credential_file)
            logger.warn(warnmsg)

        return

    def _validate_credentials(self, cred_list):
        errors = []
        warnings = []

        identifier_set = set()

        for cinfo in cred_list:
            if "identifier" in cinfo:
                identifier = cinfo["identifier"]
                if identifier in identifier_set:
                    errmsg = "Duplicate identifer found. identifier=%s" % identifier
                    errors.append(errmsg)
                else:
                    identifier_set.add(identifier)
            else:
                errmsg = "All credentials must have an identifier field. cinfo=%r" % cinfo
                errors.append(errmsg)

            if "category" in cinfo:
                category = cinfo["category"]
                if category == "basic":
                    child_errors, child_warnings =  self._validate_credential_basic(cinfo)
                    errors.extend(child_errors)
                    warnings.extend(child_warnings)
                elif category == "ssh":
                    child_errors, child_warnings =  self._validate_credential_ssh(cinfo)
                    errors.extend(child_errors)
                    warnings.extend(child_warnings)
                else:
                    warnmsg = "Unknown credential category=%s. info=%r" % (category, cinfo)
                    warnings.append(warnmsg)
            else:
                errmsg = "Credential info has no category. info=%r" % cinfo
                errors.append(errmsg)

        return errors, warnings

    def _validate_credential_basic(self, cred):
        """
            Validates the non-common fields of a 'basic' credential.
        """
        errors = []
        warnings = []

        if "username" in cred:
            if len(cred["username"].strip()) == 0:
                errmsg = "The 'username' for a basic credential cannot be empty."
                errors.append(errmsg)
        else:
            errmsg = "Basic credentials must have a 'username' field."
            errors.append(errmsg)

        if "password" not in cred:
            errmsg = "Basic credentials must have a 'password' field."
            errors.append(errmsg)

        return errors, warnings

    def _validate_credential_ssh(self, cred):
        """
            Validates the non-common fields of an 'ssh' credential.
        """
        """
        -   "identifier": "some-node"
            "category": "ssh"
            "username": "ubuntu"
            "password": "blahblah"
            "keyfile": "~/.ssh/id_somenode_rsa"

        """
        errors = []
        warnings = []

        if "username" in cred:
            if len(cred["username"].strip()) == 0:
                errmsg = "The 'username' for an SSH credential cannot be empty."
                errors.append(errmsg)
        else:
            errmsg = "SSH credentials must have a 'username' field."
            errors.append(errmsg)

        if "password" not in cred and "keyfile" not in cred:
            errmsg = "SSH credentials must have a 'password' or 'keyfile' field."
            errors.append(errmsg)
        elif "keyfile" in cred:
            keyfile = os.path.abspath(os.path.expanduser(os.path.expandvars(cred["keyfile"])))
            if not os.path.exists(keyfile):
                errmsg = "The specified SSH keyfile does not exist. file=%s" % keyfile
                errors.append(errmsg)

        return errors, warnings



__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import Optional

import errno
import socket
import struct

from akit.exceptions import AKitRuntimeError
from akit.networking.constants import IPPROTO_IPV6, SO_RECV_ANYIF

def create_unicast_socket(target_addr: str, port: int, family: socket.AddressFamily = socket.AF_INET,
    ttl: Optional[int] = None, loop: Optional[int] = None, timeout: Optional[float] = None, apple_p2p: bool = False) -> socket.socket:
    """
        Create a socket for listening for sending unicast packets to the specified target ip address.

        :param target_addr: The unicast network address the socket will be used to communicate with.
        :param port: The port to bind the socket with.
        :param family: The internet address family for the socket, either (socket.AF_INET or socket.AF_INET6)
        :param ttl: The time to live that will be attached to the packets sent by this socket.
                    0 = same host
                    1 = same subnet
                    32 = same site
                    64 = same region
                    128 = same continent
                    255 = unrestricted scope
        :param loop: Boolean value indicating if the loopback address should be included for traffic.
        :param timeout: The socket timeout to assign to the socket
        :param apple_p2p: A boolean value indicating if the socket option for Apple Peer-2-Peer should be set.

    """

    bind_addr = '0.0.0.0'
    sock = None

    if family == socket.AF_INET:
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    elif family == socket.AF_INET6:
        sock = socket.socket(socket.AF_INET6, socket.SOCK_DGRAM)
    else:
        raise AKitRuntimeError("Socket family not supported. family=%r" % family) from None

    # We need to set SO_REUSEADDR to allow the re-use of addresses by multiple processes.  This allows
    # more than one application to listen on multi-cast port addresses that are designated for specific
    # protocols.
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

    # SO_REUSEADDR should be equivalent to SO_REUSEPORT for multicast UDP sockets
    # (p 731, "TCP/IP Illustrated, Volume 2").  Some BSD-derived systems require
    # SO_REUSEPORT to be specified explicitly.  Also, not all versions of Python
    # have SO_REUSEPORT available.
    # Catch OSError for when an OS is encountered that does not support SO_REUSEPORT support.
    if hasattr(socket, "SO_REUSEPORT"):
        try:
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
        except OSError as os_err:
            if not os_err.errno == errno.ENOPROTOOPT:
                err_msg = "Error attempting to set socket option 'SO_REUSEPORT'. errno=%d" % os_err.errno
                raise AKitRuntimeError(err_msg) from os_err

    # Set the IP protocol level socket opition IP_MULTICAST_IF which is used to bind
    # the socket to the address of a specific interface for OUTBOUND multi-cast traffic
    sock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_IF, socket.inet_aton(bind_addr))

    if ttl is not None:
        if family == socket.AF_INET:
            ttl = struct.pack(b'b', ttl)
            sock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, ttl)
        else:
            sock.setsockopt(IPPROTO_IPV6, socket.IPV6_MULTICAST_HOPS, ttl)

    if loop is not None:
        if family == socket.AF_INET:
            loop = struct.pack(b'b', loop)
            sock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_LOOP, loop)
        else:
            sock.setsockopt(IPPROTO_IPV6, socket.IPV6_MULTICAST_LOOP, loop)

    if apple_p2p:
        sock.setsockopt(socket.SOL_SOCKET, SO_RECV_ANYIF, 1)

    if timeout is not None:
        sock.settimeout(timeout)

    sock.bind((bind_addr, port))

    # We also need to tell the Kernel to bind the INBOUND traffic destined for the multi-cast
    # group to the address for this interface, so we receive responses
    #member_in = socket.inet_aton(multicast_addr) + socket.inet_aton(bind_addr)
    #sock.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, member_in)

    return sock




"""
.. module:: entrypoints
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: A set of standaridized entry point functions that provide standardized test environment
               startup and test run commencement utilizing the :class:`akit.testing.unittest.testsequencer.TestSequencer`
               object.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"


import argparse
import inspect
import os
import sys


from akit.environment.context import Context
from akit.environment.variables import LOG_LEVEL_NAMES

from akit.compat import import_by_name
from akit.paths import get_path_for_testresults
from akit.testing.utilities import find_testmodule_root, find_testmodule_fullname
from akit.testing.testplus.testjob import DefaultTestJob
from akit.testing.testplus.registration.resourceregistry import ResourceRegistry
from akit.xlogging.foundations import logging_initialize, getAutomatonKitLogger


logger = getAutomatonKitLogger()

def generic_test_entrypoint():
    """
        This is the generic test entry point for test modules.  It provides a standardized set of
        commanline parameters that can be used to run test files as scripts.

    .. note::
       The `generic_test_entrypoint` is a useful tool to place at the bottom of test files to allow
       them to easily be run for debugging purposes.
    """
    import akit.activation.testrun

    # We must exit with a result code, initialize it to 0 here
    result_code = 0

    base_parser = argparse.ArgumentParser()

    base_parser.add_argument("-i", "--include", dest="includes", action="append", default=[], help="Add an include search statement.")
    base_parser.add_argument("-x", "--exclude", dest="excludes", action="append", default=[], help="Add an exclude filter statement.")
    base_parser.add_argument("--console-level", dest="consolelevel", action="store", default="INFO", choices=LOG_LEVEL_NAMES, help="The logging level for console output.")
    base_parser.add_argument("--logfile-level", dest="logfilelevel", action="store", default="DEBUG", choices=LOG_LEVEL_NAMES, help="The logging level for logfile output.")

    test_module = sys.modules["__main__"]

    ctx = Context()
    env = ctx.lookup("/environment")

    # Set the jobtype
    env["jobtype"] = "testrun"

    test_results_dir = get_path_for_testresults()
    if not os.path.exists(test_results_dir):
        os.makedirs(test_results_dir)
    env["output_directory"] = test_results_dir

    testroot = find_testmodule_root(test_module)
    module_fullname = find_testmodule_fullname(test_module, testroot=testroot)
    
    module_lastdot = module_fullname.rfind('.')
    if module_lastdot > -1:
        # We need to import every module up to the test module
        parent_module_name = module_fullname[:module_lastdot]
        import_by_name(parent_module_name)

    # Copy the test module to the name of the module_fullname name so the loader won't reload it
    sys.modules[module_fullname] = test_module


    if test_module.__name__ == "__main__":
        test_module.__name__ = module_fullname

        # Re-map the object classes from the module over to the module name we just registered the test
        # module under.
        test_class_coll = inspect.getmembers(test_module, inspect.isclass)
        for testclass_name, testclass_obj in test_class_coll:
            tcobj_module_name = testclass_obj.__module__
            if tcobj_module_name == "__main__":
                testclass_obj.__module__ = module_fullname
        
        # Re-map the object function objects from the module over to the module name we just registered
        # the test module under.
        test_func_coll = inspect.getmembers(test_module, inspect.isfunction)
        for testfunc_name, testfunc_obj in test_func_coll:
            tfobj_module_name = testfunc_obj.__module__
            if tfobj_module_name == "__main__":
                testfunc_obj.__module__ = module_fullname

        resource_registry = ResourceRegistry()
        resource_registry.rename_resource_origins_from_main(module_fullname)

    args = base_parser.parse_args()

    logging_initialize()

    includes = args.includes
    excludes = args.excludes
    if len(includes) == 0:
        includes.append("*")

    result_code = 0
    with DefaultTestJob(logger, testroot, includes=includes, excludes=excludes, test_module=test_module) as tjob:
        result_code = tjob.execute()

    sys.exit(result_code)

    return



"""

    NOTE: This is a code generated file.  This file should not be edited directly.
"""



from akit.aspects import AspectsUPnP, DEFAULT_UPNP_ASPECTS

from akit.extensible import LoadableExtension
from akit.interop.upnp.services.upnpserviceproxy import UpnpServiceProxy

class HouseStatus1ServiceProxy(UpnpServiceProxy, LoadableExtension):
    """
        This is a code generated proxy class to the 'urn:schemas-upnp-org:service:HouseStatus:1' service.
    """

    SERVICE_MANUFACTURER = 'UPnP'
    SERVICE_TYPE = 'HouseStatus1'

    SERVICE_DEFAULT_VARIABLES = {}

    SERVICE_EVENT_VARIABLES = {
        "ActivityLevel": { "data_type": "string", "default": "Regular", "allowed_list": "['Regular', 'Asleep', 'HighActivity']"},
        "DormancyLevel": { "data_type": "string", "default": "Regular", "allowed_list": "['Regular', 'Vacation', 'PetsAtHome']"},
        "OccupancyState": { "data_type": "string", "default": "Occupied", "allowed_list": "['Occupied', 'Unoccupied', 'Indeterminate']"},
    }

    def action_GetActivityLevel(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetActivityLevel action.

            :returns: "CurrentActivityLevel"
        """
        arguments = { }

        out_params = self.call_action("GetActivityLevel", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentActivityLevel",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetDormancyLevel(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetDormancyLevel action.

            :returns: "CurrentDormancyLevel"
        """
        arguments = { }

        out_params = self.call_action("GetDormancyLevel", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentDormancyLevel",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetOccupancyState(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetOccupancyState action.

            :returns: "CurrentOccupancyState"
        """
        arguments = { }

        out_params = self.call_action("GetOccupancyState", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentOccupancyState",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_SetActivityLevel(self, NewActivityLevel, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetActivityLevel action.
        """
        arguments = {
            "NewActivityLevel": NewActivityLevel,
        }

        self.call_action("SetActivityLevel", arguments=arguments, aspects=aspects)

        return

    def action_SetDormancyLevel(self, NewDormancyLevel, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetDormancyLevel action.
        """
        arguments = {
            "NewDormancyLevel": NewDormancyLevel,
        }

        self.call_action("SetDormancyLevel", arguments=arguments, aspects=aspects)

        return

    def action_SetOccupancyState(self, NewOccupancyState, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetOccupancyState action.
        """
        arguments = {
            "NewOccupancyState": NewOccupancyState,
        }

        self.call_action("SetOccupancyState", arguments=arguments, aspects=aspects)

        return



"""

    NOTE: This is a code generated file.  This file should not be edited directly.
"""



from akit.extensible import LoadableExtension
from akit.interop.upnp.services.upnpserviceproxy import UpnpServiceProxy

class RenderingControl1ServiceProxy(UpnpServiceProxy, LoadableExtension):
    """
        This is a code generated proxy class to the 'RenderingControl1' service.
    """

    SERVICE_MANUFACTURER = 'MicrosoftCorporation'
    SERVICE_TYPE = 'urn:schemas-upnp-org:service:RenderingControl:1'

    SERVICE_EVENT_VARIABLES = {
        "LastChange": { "data_type": "string", "default": None, "allowed_list": None},
    }

    def action_GetMute(self, InstanceID, Channel, extract_returns=True):
        """
            Calls the GetMute action.

            :returns: "CurrentMute"
        """
        arguments = {
            "InstanceID": InstanceID,
            "Channel": Channel,
        }

        out_params = self._proxy_call_action("GetMute", arguments=arguments)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentMute",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetVolume(self, InstanceID, Channel, extract_returns=True):
        """
            Calls the GetVolume action.

            :returns: "CurrentVolume"
        """
        arguments = {
            "InstanceID": InstanceID,
            "Channel": Channel,
        }

        out_params = self._proxy_call_action("GetVolume", arguments=arguments)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentVolume",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_ListPresets(self, InstanceID, extract_returns=True):
        """
            Calls the ListPresets action.

            :returns: "CurrentPresetNameList"
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        out_params = self._proxy_call_action("ListPresets", arguments=arguments)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentPresetNameList",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_SelectPreset(self, InstanceID, PresetName, extract_returns=True):
        """
            Calls the SelectPreset action.

            :returns: "result"
        """
        arguments = {
            "InstanceID": InstanceID,
            "PresetName": PresetName,
        }

        out_params = self._proxy_call_action("SelectPreset", arguments=arguments)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("result",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_SetMute(self, InstanceID, Channel, DesiredMute, extract_returns=True):
        """
            Calls the SetMute action.

            :returns: "result"
        """
        arguments = {
            "InstanceID": InstanceID,
            "Channel": Channel,
            "DesiredMute": DesiredMute,
        }

        out_params = self._proxy_call_action("SetMute", arguments=arguments)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("result",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_SetVolume(self, InstanceID, Channel, DesiredVolume, extract_returns=True):
        """
            Calls the SetVolume action.

            :returns: "result"
        """
        arguments = {
            "InstanceID": InstanceID,
            "Channel": Channel,
            "DesiredVolume": DesiredVolume,
        }

        out_params = self._proxy_call_action("SetVolume", arguments=arguments)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("result",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args



"""
.. module:: execution
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: A module that provides the function which implemented the execution workflow
               of an individual workpacket.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

import os

from akit.compat import import_by_name

from akit.exceptions import AKitConfigurationError


class JumpContext:
    def __init__(self) -> None:
        self.on_failure_section = None
        self.on_success_section = None


def execute_tasklist(logger, jump_context: JumpContext, parameters: dict, tasklist: list, ordinal_prefix: str=None, **kwargs):

    result_code = 0

    for index, item in enumerate(tasklist):
 
        nxt_ordinal = str(index)
        if ordinal_prefix is not None:
            nxt_ordinal = "{}.{}".format(ordinal_prefix, nxt_ordinal)

        if "task" in item:

            task_info: dict = item
            task_label = task_info["task"]

            if jump_context.on_success_section is not None:
                if jump_context.on_success_section != task_label:
                    continue
                else:
                    jump_context.on_success_section = None

            if jump_context.on_failure_section is not None:
                if jump_context.on_failure_section != task_label:
                    continue
                else:
                    jump_context.on_failure_section = None

            task_type = item["tasktype"]

            task_module_name, task_module_class = task_type.split("@")
            task_module = import_by_name(task_module_name)

            if hasattr(task_module, task_module_class):
                task_class = getattr(task_module, task_module_class)

                task_instance = task_class(nxt_ordinal, task_label, task_info, logger)

                task_result = task_instance.execute(parameters=parameters, **kwargs)
                if task_result != 0:
                    result_code = 1
                    if task_instance.onfailure is not None:
                        failure_section = task_instance.onfailure
                elif task_instance.onsuccess is not None:
                    jump_context.on_success_section = task_instance.onsuccess

            else:
                error_msg = "The specified task module %r does not contain a class %r" % (
                    task_module_name, task_module_class)
                raise AKitConfigurationError(error_msg) from None

        elif "taskgroup" in item:
            
            tg_info: dict = item
            tg_label = tg_info["taskgroup"]

            if jump_context.on_success_section is not None:
                if jump_context.on_success_section != task_label:
                    continue
                else:
                    jump_context.on_success_section = None

            if jump_context.on_failure_section is not None:
                if jump_context.on_failure_section != task_label:
                    continue
                else:
                    jump_context.on_failure_section = None

            group_prefix = nxt_ordinal
            if ordinal_prefix is not None:
                nxt_ordinal = "{}.{}".format(ordinal_prefix, group_prefix)

            tg_items = tg_info["items"]

            tg_result = execute_tasklist(logger, jump_context, parameters, tg_items, ordinal_prefix=group_prefix)
            if tg_result != 0:
                result_code = 1
                if task_instance.onfailure is not None:
                    jump_context.on_failure_section = task_instance.onfailure
            elif task_instance.onsuccess is not None:
                jump_context.on_success_section = task_instance.onsuccess

        else:
            pass

    return result_code


def execute_workflow(logger, *, environment: dict, parameters: dict, workflow: list, **kwargs):

    # Publish the environment variables so they will take effect in the current
    # process and any sub-processes lauched from this process
    for key, val in environment.items():
        os.environ[key] = val

    result_code = 0
    jump_context = JumpContext()

    execute_tasklist(logger, jump_context, parameters, workflow)

    return result_code



"""
.. module:: topologydescription
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Module containing the :class:`TopologyDescription` class.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>

"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import List, Optional, Tuple, TYPE_CHECKING

import os
import shutil
import traceback
import yaml

from akit.exceptions import AKitConfigurationError, AKitRuntimeError
from akit.xlogging.foundations import getAutomatonKitLogger

class TopologyDescription:
    """
        The base class for all derived :class:`TopologyDescription` objects.  The
        :class:`TopologyDescription` is used to load a description of the overlayed
        relationships between the entities in the test landscape.
    """

    def load(self, topology_file: str, log_to_directory: Optional[str]=None):
        """
            Loads and validates the topology description file.
        """
        logger = getAutomatonKitLogger()

        topology_info = None


        with open(topology_file, 'r') as tf:
            tfcontent = tf.read()
            topology_info = yaml.safe_load(tfcontent)

        if log_to_directory is not None:
            try:
                topology_file_basename = os.path.basename(topology_file)
                topology_file_basename, topology_file_ext = os.path.splitext(topology_file_basename)

                topology_file_copy = os.path.join(log_to_directory, "topology-declared{}".format(topology_file_ext))
                shutil.copy2(topology_file, topology_file_copy)
            except Exception as xcpt:
                err_msg = "Error while logging the topology file (%s)%s%s" % (
                    topology_file, os.linesep, traceback.format_exc())
                raise AKitRuntimeError(err_msg) from xcpt

        errors, warnings = self.validate_topology(topology_info)

        if len(errors) > 0:
            errmsg_lines = [
                "ERROR Topology validation failures:"
            ]
            for err in errors:
                errmsg_lines.append("    %s" % err)

            errmsg = os.linesep.join(errmsg_lines)
            raise AKitConfigurationError(errmsg) from None

        if len(warnings) > 0:
            for wrn in warnings:
                logger.warn("Topology Configuration Warning: (%s)" % wrn)

        return topology_info

    def validate_topology(self, topology_info) -> Tuple[List[str], List[str]]:
        """
            Overloaded by derived topology classes in order to validate
            customized device topology description files.
        """
        errors = []
        warnings = []

        return errors, warnings




__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import TYPE_CHECKING

from threading import Lock

from akit.extensionpoints import AKitExtensionPoints

if TYPE_CHECKING:
    from akit.interop.landscaping.landscape import Landscape

LANDSCAPE_SINGLETON = None

SINGLETON_LOCK = Lock()

def LandscapeSingleton() -> "Landscape":

    global SINGLETON_LOCK
    global LANDSCAPE_SINGLETON

    # If the singleton is already set, don't bother grabbing a lock
    # to set it.  The full path of the setting of the singleton will only
    # ever be taken once
    if LANDSCAPE_SINGLETON is None:
        SINGLETON_LOCK.acquire()
        try:
            if LANDSCAPE_SINGLETON is None:
                LandscapeType = AKitExtensionPoints().get_landscape_type()
                LANDSCAPE_SINGLETON = LandscapeType()
        finally:
            SINGLETON_LOCK.release()

    return LANDSCAPE_SINGLETON


__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import Union

import inspect
import os
import fnmatch
import sys
import traceback

from akit.compat import import_by_name
from akit.exceptions import AKitSemanticError

from akit.testing.testplus.testref import TestRef

from akit.xlogging.foundations import getAutomatonKitLogger

logger = getAutomatonKitLogger()

def collect_test_references(root, included_files, filter_package, filter_module, filter_testname, test_prefix):
    """
        Finds the test references based on the expression provided and the excludes
        for this class.  The `find_test_references` method is intended to be called
        multiple times, once with each include expression provided by the users.

        :param expression: An include expression to process and collect references for.
    """

    test_references = {}

    import_errors = {}

    root_pkgname = os.path.basename(root)

    # Go through the files and import them, then go through the classes and find the TestPack and
    # TestContainer objects that match the specified include expression criteria
    rootlen = len(root)
    for ifile in included_files:
        modname = None
        try:
            ifilebase, _ = os.path.splitext(ifile)
            ifileleaf = ifilebase[rootlen:].strip("/")
            modname = "{}.{}".format(root_pkgname, ifileleaf.replace("/", "."))

            # Import the module for the file being processed
            mod = None
            if modname in sys.modules:
                mod = sys.modules[modname]
            else:
                mod = import_by_name(modname)

            # Go through all of the members of the
            candidate_function_coll = inspect.getmembers(mod, inspect.isfunction)
            for function_name, function_obj in candidate_function_coll:
                cand_module_name = function_obj.__module__
                # We only want to include the classes that are from the target module
                if cand_module_name != modname:
                    continue

                if function_name.startswith(test_prefix):
                    if filter_testname is not None:
                        # If we have a testname expression only add a reference to the test function
                        # if the function_name matches the filter expression
                        if fnmatch.fnmatch(function_name, filter_testname):
                            tref = TestRef(function_obj)
                            test_references[tref.test_name] = tref
                    else:
                        tref = TestRef(function_obj)
                        test_references[tref.test_name] = tref

        except ImportError:
            errmsg = traceback.format_exc()
            print(errmsg)
            import_errors[ifile] = (modname, ifile, errmsg)

    import_errors.update(import_errors)

    for modname, ifile, errmsg in import_errors.values():
        logger.error("TestCase: Import error filename=%r" % ifile)

    return test_references, import_errors






from akit.xlogging.foundations import getAutomatonKitLogger

from akit.exceptions import akit_skip

from .markers import (
    mark_categories,
    mark_keywords,
    mark_priority,
    mark_descendent_categories,
    mark_descendent_keywords,
    mark_descendent_priority
)

from .parameters import (
    originate_parameter,
    param
)


from .resources import (
    integration,
    resource,
    scope
)

from .verification import (
    assert_dict_response_has_keys,
    assert_dict_response_has_paths,
    assert_equal,
    assert_expression,
    assert_greater,
    assert_lessthan,
    assert_list_length,
    assert_list_length_greater,
    assert_type
)

logger = getAutomatonKitLogger()

skip_test = akit_skip

__all__ = [
    assert_dict_response_has_keys,
    assert_dict_response_has_paths,
    assert_equal,
    assert_expression,
    assert_greater,
    assert_lessthan,
    assert_list_length,
    assert_list_length_greater,
    assert_type,
    integration,
    logger,
    mark_categories,
    mark_keywords,
    mark_priority,
    mark_descendent_categories,
    mark_descendent_keywords,
    mark_descendent_priority,
    originate_parameter,
    param,
    resource,
    scope,
    skip_test
]



__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"


import os

from datetime import datetime

import click

from akit.xtime import FORMAT_DATETIME

HELP_RUNTIME = "The full path of the runtime file to use for the creation of the output folder path."

@click.group("utilities", help="Contains utility commands for use in integration scripts.")
def group_akit_utilities():
    return

@click.command("outputfolder", help="Creates the full path of a time-stamped output folder based a <starttime> parameter.")
@click.option("--runtime", "runtime_file", default=None, required=False, help=HELP_RUNTIME)
@click.argument("starttime")
def command_akit_utilities_outputfolder(runtime_file, starttime):
    
    # IMPORTANT: We need to load the context first because it will trigger the loading
    # of the default user configuration
    from akit.environment.context import Context

    from akit.environment.variables import JOB_TYPES
    from akit.environment.optionoverrides import override_starttime, override_config_runtime

    ctx = Context()
    env = ctx.lookup("/environment")

    # We need to set the job type before we trigger activation.
    env["jobtype"] = JOB_TYPES.CONSOLE

    import akit.activation.console

    if runtime_file is not None:
        override_config_runtime(runtime_file)

    if starttime is not None:
        override_starttime(starttime)

    from akit.paths import get_path_for_testresults

    ts_string = get_path_for_testresults()

    print(ts_string)

    return

@click.command("timestamp", help="Generates a timestamp from the current time.")
def command_akit_utilities_timestamp():

    timestamp = datetime.now()
    ts_string = str(timestamp).replace(" ", "T")

    print(ts_string)

    return

group_akit_utilities.add_command(command_akit_utilities_outputfolder)
group_akit_utilities.add_command(command_akit_utilities_timestamp)



"""
.. module:: exceptions
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Module containing the exceptions that are raised by the code in the Automation Kit

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

import inspect
import os
import traceback

from akit.xinspect import get_caller_function_name
from akit.xformatting import split_and_indent_lines


MEMBER_TRACE_POLICY = "__traceback_format_policy__"


class TracebackFormatPolicy:
    Brief = "Brief"
    Full = "Full"
    Hide = "Hide"


VALID_MEMBER_TRACE_POLICY = ["Brief", "Full", "Hide"]


__traceback_format_policy__ = TracebackFormatPolicy.Hide


class TRACEBACK_CONFIG:
    TRACEBACK_POLICY_OVERRIDE = None
    TRACEBACK_MAX_FULL_DISPLAY = 10


def akit_assert(eresult, errmsg, from_exception=None):
    if not eresult:
        raise AKitAssertionError(errmsg) from from_exception


def akit_skip(skipmsg):
    raise AKitSkipError(skipmsg)


def collect_stack_frames(calling_frame, ex_inst):

    global TRACEBACK_POLICY_OVERRIDE

    max_full_display = TRACEBACK_CONFIG.TRACEBACK_MAX_FULL_DISPLAY

    last_items = None
    tb_code = None
    tb_lineno = None

    tb_frames = []

    for tb_frame, tb_lineno in traceback.walk_stack(calling_frame):
        tb_frames.insert(0, (tb_frame, tb_lineno))
        if tb_frame.f_code.co_name == '<module>':
            break 

    for tb_frame, tb_lineno in traceback.walk_tb(ex_inst.__traceback__):
        tb_frames.append((tb_frame, tb_lineno))

    tb_frames.reverse()

    traceback_list = []

    for tb_frame, tb_lineno in tb_frames:
        tb_code = tb_frame.f_code
        co_filename = tb_code.co_filename
        co_name = tb_code.co_name
        co_arg_names = tb_code.co_varnames[:tb_code.co_argcount]
        co_argcount = tb_code.co_argcount
        co_locals = tb_frame.f_locals

        co_format_policy = TracebackFormatPolicy.Full

        if TRACEBACK_CONFIG.TRACEBACK_POLICY_OVERRIDE is None:
            co_module = inspect.getmodule(tb_code)
            if co_module and hasattr(co_module, MEMBER_TRACE_POLICY):
                cand_format_policy = getattr(co_module, MEMBER_TRACE_POLICY)
                if cand_format_policy in VALID_MEMBER_TRACE_POLICY:
                    co_format_policy = cand_format_policy
        else:
            co_format_policy = TRACEBACK_CONFIG.TRACEBACK_POLICY_OVERRIDE

        items = [co_filename, tb_lineno, co_name, "", None]
        if last_items is not None:
            code_args = []
            for argidx in range(0, co_argcount):
                argname = co_arg_names[argidx]
                argval = co_locals[argname]
                code_args.append("%s=%r" % (argname, argval))

            last_items[-2] = "%s(%s)" % (co_name, ", ".join(code_args)) # pylint: disable=unsupported-assignment-operation

        last_items = items

        traceback_list.append(items)
        last_items = items

        if max_full_display > 0 and co_format_policy == TracebackFormatPolicy.Full \
            and os.path.exists(co_filename) and co_filename.endswith(".py"):
            context_lines, context_startline = inspect.getsourcelines(tb_code)
            context_lines = [cline.rstrip() for cline in context_lines]
            clindex = (tb_lineno - context_startline)
            last_items[-2] = context_lines[clindex].strip()
            last_items[-1] = context_lines
            max_full_display -= 1

    return traceback_list


def format_exception(ex_inst):
    exc_lines = []

    etypename = type(ex_inst).__name__
    eargs = ex_inst.args

    exmsg_lines = [
        "%s: %s" % (etypename, repr(eargs).rstrip(",")),
        "TRACEBACK (most recent call last):"
    ]

    previous_frame = inspect.currentframe().f_back

    stack_frames = collect_stack_frames(previous_frame, ex_inst)
    stack_frames_len = len(stack_frames)
    for co_filename, co_lineno, co_name, co_code, co_context in stack_frames:

        exmsg_lines.extend([
            '  File "%s", line %d, in %s' % (co_filename, co_lineno, co_name),
            "    %s" % co_code
        ])

        if hasattr(ex_inst, "context") and co_name in ex_inst.context:
            cxtinfo = ex_inst.context[co_name]
            exmsg_lines.append('    %s:' % cxtinfo["label"])
            exmsg_lines.extend(split_and_indent_lines(cxtinfo["content"], 2, indent=3))
        elif co_context is not None and len(co_context) > 0 and stack_frames_len > 1:
            exmsg_lines.append('    CONTEXT:')
            firstline = co_context[0]
            lstrip_len = len(firstline) - len(firstline.lstrip())
            co_context = [cline[lstrip_len:] for cline in co_context]
            co_context = ["      %s" % cline for cline in co_context]
            exmsg_lines.extend(co_context)
        exmsg_lines.append('')

    return exmsg_lines


class AKitErrorEnhancer:
    def __init__(self, *args, **kwargs):
        self._context = {}
        return

    @property
    def context(self):
        return self._context

    def add_context(self, content, label="CONTEXT"):
        """
            Adds context to an exception and associates it with the function context
            on the stack.
        """
        caller_name = get_caller_function_name()

        self._context[caller_name] = {
            "label": label,
            "content": content
        }

        return


class AKitBaseException(BaseException, AKitErrorEnhancer):
    """
        The base error object for Automation Kit errors that we don't want to be catchable as
        generic exceptions.  The :class:`AKitBaseException` object lets semantic errors and any
        other error that should not be caught bypass a generic exception hanlder.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        return


class AKitError(Exception, AKitErrorEnhancer):
    """
        The base error object for Automation Kit errors.  The :class:`AKitError` serves as aa base
        type and also provides some additional functionality for adding context to errors and
        formatting exception output.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        return


def add_exception_context(xcpt: BaseException, content, label="CONTEXT"):
    """
        Allows for the enhancing of Non Automation Kit exceptions.
    """

    # AKitErrorEnhancer just uses Duck typing so it should be safe to dynamically
    # append any exception that does not already inherit include AKitErrorEnhancer
    # in its base clases list.
    if AKitErrorEnhancer not in xcpt.__bases__:
        xcpt.__bases__ += (AKitErrorEnhancer,)
        xcpt._context = {}

    xcpt.add_context(content, label=label)

    return

# ==================================================================================
#                     CONFIGURATION - BASE ERROR CLASSIFICATIONS
# ==================================================================================
class AKitConfigurationError(AKitBaseException):
    """
        The base error object for errors that indicate that there is an issue related
        to improper configuration.
    """

class AKitInvalidConfigError(AKitConfigurationError):
    """
        This error is raised when an IntegrationCoupling object has been passed invalid configuration parameters.
    """

class AKitMissingConfigError(AKitConfigurationError):
    """
        This error is raised when an IntegrationCoupling object is missing required configuration parameters.
    """


# ==================================================================================
#                     IMPROPER USE - BASE ERROR CLASSIFICATIONS
# ==================================================================================
class AKitSemanticError(AKitBaseException):
    """
        The base error object for errors that indicate that there is an issue with
        a piece of automation code and with the way the Automation Kit code is being
        utilized.
    """


# ==================================================================================
#                        ENVIRONMENTAL - LANDSCAPE RELATED ERRORS
# ==================================================================================
class AKitLandscapeError(AKitError):
    """
        The base error object for errors that indicate that there is an issue related
        to the interaction, usage or consumption of an environmental resources.
    """

class AKitInitialConnectivityError(AKitLandscapeError):
    """
        This error is raised when an IntegrationCoupling object is unable to establish an initial level of
        connectivity with a connected resource.
    """

class AKitMissingResourceError(AKitLandscapeError):
    """
        This error is raised when an device or resources was declared in the landscape.yaml file
        but was not able to be found during device or resource discovery.
    """

class AKitResourceError(AKitLandscapeError):
    """
        This error is raised when a resource requirement was not able to be obtained.
    """

# ==================================================================================
#                           RUNTIME RELATED ERRORS
# ==================================================================================

class AKitRuntimeError(RuntimeError, AKitErrorEnhancer):
    """
        The base error object for errors that indicate that an error was produced during
        the execution of task or test code and the error was not able to be classified
        as Configuration, Landscape, or Semantic related.
    """
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        return

class AKitCommunicationsProtocolError(AKitRuntimeError):
    """
        This is the base error for exceptions that are related to communciations protocols
    """

class AKitEstablishPresenceError(AKitLandscapeError):
    """
        This error occurs when an integration coupling has trouble establishing a presence in the
        test landscape.
    """

class AKitCommandError(AKitCommunicationsProtocolError):
    """
        This error is the base error for HTTP requests based errors.
    """
    def __init__(self, message, status, stdout, stderr, *args, **kwargs):
        super().__init__(message, *args, **kwargs)
        self.status = status
        self.stdout = stdout
        self.stderr = stderr
        return

class AKitHTTPRequestError(AKitCommunicationsProtocolError):
    """
        This error is the base error for HTTP requests based errors.
    """
    def __init__(self, message, requrl, status_code, reason, *args, **kwargs):
        super().__init__(message, *args, **kwargs)
        self.requrl = requrl
        self.status_code = status_code
        self.reason = reason
        return

class AKitOpenWRTRequestError(AKitCommunicationsProtocolError):
    """
        This error is the base error OpenWRT inter-op command request errors.
    """

class AKitOutOfScopeError(AKitRuntimeError):
    """
        This error is raised when a method is called on a ScopeCoupling that is not in scope.  A test can have,
        multiple ScopeCoupling(s) and can run in multiple scopes but the test must be instantiated and run in
        each scope individually.
    """

class AKitReaderError(AKitRuntimeError):
    """
        This error indicates that an error was encoutered while attempting to read and extract data
        from text content.
    """

class AKitRequestStopError(AKitRuntimeError):
    """
        This error is raised when a test indicates it wants to stop an automation run.  The `TestSequencer`
        may or may not stop the automation run as a result of a test or scope raising this error.  The
        `TestSequencer` looks at the current runtime context which was set by the commandline arguements
        and will stop the test run if the runtime context indicates that stopping is allowed.
    """

class AKitScopeEntryError(AKitRuntimeError):
    """
        This error is raised when a ScopeCoupling was unable to complete the entry of a scope.
    """

class AKitSetupError(AKitRuntimeError):
    """
        An error occured during the setup of a Task, Test, Step or Process
    """

class AKitSkipError(AKitRuntimeError):
    """
        This error is raised when a test indicates it wants to be skipped while being run
    """
    def __init__(self, *args, reason=None, bug=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.reason = reason
        self.bug = bug
        return

class AKitLooperError(AKitRuntimeError):
    """
        This error is raised when an error occurs with the use of the :class:`LooperPool` or
        :class:`Looper` objects.
    """

class AKitLooperQueueShutdownError(AKitRuntimeError):
    """
        This error is raised when work is being queued on a :class:`LooperQueue` thaat has
        been shutdown and when a worker thread is attempting to wait for work on an empty
        queue.
    """

class AKitUnknownParameterError(AKitRuntimeError):
    """
        This error is raised when the test framework encounters an unknown or unresolvable parameter.
    """

class AKitAbstractMethodError(AKitRuntimeError):
    """
        This error is raised when an abstract method has been called.
    """

class AKitNotImplementedError(NotImplementedError, AKitErrorEnhancer):
    """
        This error is raised when a method is called that has not yet been implemented.
    """

class AKitNotOverloadedError(AKitRuntimeError):
    """
        This error is raised when a method that must be overloaded has not been overridden.
    """
class AKitNotSupportedError(AKitRuntimeError):
    """
        This error is raised when a method that must be overloaded has not been overridden.
    """

class AKitRecursionError(RecursionError, AKitErrorEnhancer):
    """
        This error is raised when a method that must be overloaded has not been overridden.
    """
    def __init__(self, *args, reason=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.reason = reason
        return

class AKitServiceUnAvailableError(AKitRuntimeError):
    """
        This error is raised when a service is requested that is not currently available.
    """

# ==================================================================================
#                           TOOLING ERRORS
# ==================================================================================

class AKitGeneratorError(AKitError):
    """
        This error is raised when a code generator tool has encountered an issue.
    """

class AKitGenerateItemError(AKitError):
    """
        This error is raised when a code generator tool has missing data or insuffient
        data to continue to generate the current item.
    """

# ==================================================================================
#                         ENHANCED BUILTIN ERRORS
# ==================================================================================

class AKitArithmeticError(ArithmeticError, AKitErrorEnhancer):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        return

class AKitFloatingPointError(FloatingPointError, AKitErrorEnhancer):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        return

class AKitOverflowError(OverflowError, AKitErrorEnhancer):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        return

class AKitZeroDivisionError(ZeroDivisionError, AKitErrorEnhancer):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        return

class AKitAssertionError(AssertionError, AKitErrorEnhancer):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        return
class AKitAttributeError(AttributeError, AKitErrorEnhancer):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        return

class AKitBufferError(BufferError, AKitErrorEnhancer):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        return

class AKitEOFError(EOFError, AKitErrorEnhancer):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        return

class AKitImportError(ImportError, AKitErrorEnhancer):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        return

class AKitModuleNotFoundError(ModuleNotFoundError, AKitErrorEnhancer):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        return

class AKitLookupError(LookupError, AKitErrorEnhancer):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        return

class AKitIndexError(LookupError, AKitErrorEnhancer):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        return

class AKitKeyError(KeyError, AKitErrorEnhancer):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        return

class AKitMemoryError(MemoryError, AKitErrorEnhancer):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        return

class AKitNameError(NameError, AKitErrorEnhancer):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        return

class AKitUnboundLocalError(UnboundLocalError, AKitErrorEnhancer):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        return

class AKitOSError(OSError, AKitErrorEnhancer):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        return

class AKitBlockingIOError(BlockingIOError, AKitErrorEnhancer):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        return

class AKitChildProcessError(ChildProcessError, AKitErrorEnhancer):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        return

class AKitConnectionError(ConnectionError, AKitErrorEnhancer):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        return

class AKitBrokenPipeError(BrokenPipeError, AKitErrorEnhancer):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        return

class AKitConnectionAbortedError(ConnectionAbortedError, AKitErrorEnhancer):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        return

class AKitConnectionRefusedError(ConnectionRefusedError, AKitErrorEnhancer):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        return

class AKitConnectionResetError(ConnectionResetError, AKitErrorEnhancer):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        return

class AKitFileExistsError(FileExistsError, AKitErrorEnhancer):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        return

class AKitFileNotFoundError(FileNotFoundError, AKitErrorEnhancer):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        return

class AKitInterruptedError(InterruptedError, AKitErrorEnhancer):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        return

class AKitIsADirectoryError(IsADirectoryError, AKitErrorEnhancer):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        return

class AKitNotADirectoryError(NotADirectoryError, AKitErrorEnhancer):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        return

class AKitPermissionError(PermissionError, AKitErrorEnhancer):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        return

class AKitProcessLookupError(ProcessLookupError, AKitErrorEnhancer):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        return

class AKitTimeoutError(TimeoutError, AKitErrorEnhancer):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        return

class AKitReferenceError(ReferenceError, AKitErrorEnhancer):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        return

class AKitSyntaxError(SyntaxError, AKitErrorEnhancer):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        return

class AKitIndentationError(IndentationError, AKitErrorEnhancer):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        return

class AKitTabError(TabError, AKitErrorEnhancer):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        return

class AKitSystemError(SystemError, AKitErrorEnhancer):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        return

class AKitTypeError(TypeError, AKitErrorEnhancer):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        return

class AKitValueError(ValueError, AKitErrorEnhancer):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        return

class AKitUnicodeError(UnicodeError, AKitErrorEnhancer):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        return

class AKitUnicodeDecodeError(UnicodeDecodeError, AKitErrorEnhancer):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        return

class AKitUnicodeEncodeError(UnicodeEncodeError, AKitErrorEnhancer):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        return

class AKitUnicodeTranslateError(UnicodeTranslateError, AKitErrorEnhancer):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        return



__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import Optional

class BaseCredential:
    """
        The :class:`BaseCredential` is the base container object for credentials passed in the landscape
        configuration file.

        .. code:: yaml
            "identifier": "player-ssh"
            "category": "(category)"
    """
    def __init__(self, *, identifier: str, category: str, role: Optional[str]="priv"):
        """
            :param identifier: The identifier that is used to reference this credential.  (required)
            :param category: The category of credential.
            :param role: Identifies the role of the credential
        """
        if len(identifier) == 0:
            raise ValueError("The MuseCredential constructor requires a 'identifier' parameter be provided.")
        if len(category) == 0:
            raise ValueError("The MuseCredential constructor requires a 'category' parameter be provided.")

        self._identifier = identifier
        self._category = category
        self._role = role
        return

    @property
    def category(self):
        return self._category

    @property
    def identifier(self):
        return self._identifier

    @property
    def role(self):
        return self._role



__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import Optional

import os

from akit.exceptions import AKitConfigurationError
from akit.interop.credentials.basecredential import BaseCredential

class BasicCredential(BaseCredential):
    """
        The :class:`BasicCredential` is a container object for a basic username and password based credential.

        .. code:: yaml
            "identifier": "basic-login"
            "category": "basic"
            "username": "ubuntu"
            "password": "@@&_@@&_LetMeComeIn"

    """

    def __init__(self, *, identifier: str, category: str, username: str, password: str, role: Optional[str] = "priv"):
        """
            :param identifier: The identifier that is used to reference this credential.  (required)
            :param category: The category of credential.
            :param username: The username associated with this credential.
            :param password: The password associated with this credential.  A password is not required if a
                             keyfile parameter is provided or if 'allow_agent' is passed as 'True'.
        """
        super().__init__(identifier=identifier, category=category, role=role)

        if category != "basic":
            raise ValueError("The BasicCredential should only be given credentials of category 'basic'.")
        if len(username) == 0:
            raise ValueError("The BasicCredential constructor requires a 'username' parameter be provided.")
        if len(password) == 0:
            raise ValueError("The BasicCredential constructor requires one of: 'password is not None'.")

        self._username = username
        self._password = password
        return

    @property
    def password(self):
        return self._password

    @property
    def username(self):
        return self._username

    @classmethod
    def validate(cls, cred_info):

        errmsg_lines = []

        if "password" not in cred_info:
                errmsg_lines.append("    * missing 'password' in basic credential.")

        if "username" not in cred_info:
                errmsg_lines.append("    * missing 'username' in basic credential.")

        if len(errmsg_lines) > 0:
            identifier = "????"
            if "identifier" in cred_info:
                identifier = cred_info["identifier"]

            errmsg = "Errors found while validating the '{}' basic credential:".format(identifier)
            errmsg_lines.insert(0, errmsg)
            errmsg = os.linesep.join(errmsg_lines)

            raise AKitConfigurationError(errmsg) from None

        return



"""
.. module:: upnperrors
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Module containing the UPNP error classes.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>

"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from enum import Enum

from akit.exceptions import AKitCommunicationsProtocolError

class UpnpError(AKitCommunicationsProtocolError):
    """
        The base class for UPnP based errors.
    """
    def __init__(self, errorCode, errorDescription, extra=None):
        self.errorCode = errorCode
        self.errorDescription = errorDescription
        self.extra = extra

        errstr = "UpnpError: errorCode=%s errorDescription=%r " % (self.errorCode, self.errorDescription)
        if self.extra is not None:
            errstr += self.extra

        super(UpnpError, self).__init__(errstr)
        return

class UpnpErrorCodes(Enum):
    """
        An enumeration of the standard UPnP error codes
    """
    InvalidAction = 401
    InvalidArgs = 402
    DepricatedA = 403
    ActionFailed = 501
    ArgumentValueInvalid = 600
    ArgumentValueOutOfRange = 601
    OptionalActionNotImplemented = 602
    OutOfMemory = 603
    HumanInterventionRequired = 604
    StringArgumentTooLong = 605
    ActionNotAuthorized = 606
    SignatureFailure = 607
    SignatureMissing = 608
    NotEncrypted = 609
    InvalidSequence = 610
    InvalidControlURL = 611
    NoSuchSession = 612


UPNP_ERROR_TEST_LOOKUP = {
    401: "Invalid Action",
    402: "Invalid Arguments",
    403: "Depricated Error Code",
    501: "Action Failed",
    600: "Argument Value Invalid",
    601: "Argument Value Out of Range",
    602: "Optional Action Not Implemented",
    603: "Out of Memory",
    604: "Human Intervention Required",
    605: "String Argument Too Long",
    606: 'Action Not Authorized',
    607: 'Signature Failure',
    608: 'Signature Missing',
    609: 'Not Encrypted',
    610: 'Invalid Sequence',
    611: 'Invalid Control URL',
    612: 'No Such Session'
}

# 606-6124 Reserved These ErrorCodes are reserved for UPnP DeviceSecurity.
# 613-699 TBD Common action errors. Defined by UPnP Forum Technical Committee.
# 700-799 TBD Action-specific errors defined by UPnP Forum working committee.
# 800-899 TBD Action-specific errors for non-standard actions. Defined by UPnP vendor.



"""

    NOTE: This is a code generated file.  This file should not be edited directly.
"""



from akit.aspects import AspectsUPnP, DEFAULT_UPNP_ASPECTS

from akit.extensible import LoadableExtension
from akit.interop.upnp.services.upnpserviceproxy import UpnpServiceProxy

class ConnectionManager3ServiceProxy(UpnpServiceProxy, LoadableExtension):
    """
        This is a code generated proxy class to the 'urn:schemas-upnp-org:service:ConnectionManager:3' service.
    """

    SERVICE_MANUFACTURER = 'UPnP'
    SERVICE_TYPE = 'ConnectionManager3'

    SERVICE_DEFAULT_VARIABLES = {
        "ClockUpdateID": { "data_type": "ui4", "default": None, "allowed_list": None},
        "FeatureList": { "data_type": "string", "default": None, "allowed_list": None},
    }

    SERVICE_EVENT_VARIABLES = {
        "CurrentConnectionIDs": { "data_type": "string", "default": None, "allowed_list": None},
        "DeviceClockInfoUpdates": { "data_type": "string", "default": None, "allowed_list": None},
        "SinkProtocolInfo": { "data_type": "string", "default": None, "allowed_list": None},
        "SourceProtocolInfo": { "data_type": "string", "default": None, "allowed_list": None},
    }

    def action_ConnectionComplete(self, ConnectionID, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the ConnectionComplete action.
        """
        arguments = {
            "ConnectionID": ConnectionID,
        }

        self.call_action("ConnectionComplete", arguments=arguments, aspects=aspects)

        return

    def action_GetCurrentConnectionIDs(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetCurrentConnectionIDs action.

            :returns: "ConnectionIDs"
        """
        arguments = { }

        out_params = self.call_action("GetCurrentConnectionIDs", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("ConnectionIDs",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetCurrentConnectionInfo(self, ConnectionID, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetCurrentConnectionInfo action.

            :returns: "RcsID", "AVTransportID", "ProtocolInfo", "PeerConnectionManager", "PeerConnectionID", "Direction", "Status"
        """
        arguments = {
            "ConnectionID": ConnectionID,
        }

        out_params = self.call_action("GetCurrentConnectionInfo", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("RcsID", "AVTransportID", "ProtocolInfo", "PeerConnectionManager", "PeerConnectionID", "Direction", "Status",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetFeatureList(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetFeatureList action.

            :returns: "FeatureList"
        """
        arguments = { }

        out_params = self.call_action("GetFeatureList", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("FeatureList",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetProtocolInfo(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetProtocolInfo action.

            :returns: "Source", "Sink"
        """
        arguments = { }

        out_params = self.call_action("GetProtocolInfo", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("Source", "Sink",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetRendererItemInfo(self, ItemInfoFilter, ItemMetadataList, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetRendererItemInfo action.

            :returns: "ItemRenderingInfoList"
        """
        arguments = {
            "ItemInfoFilter": ItemInfoFilter,
            "ItemMetadataList": ItemMetadataList,
        }

        out_params = self.call_action("GetRendererItemInfo", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("ItemRenderingInfoList",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_PrepareForConnection(self, RemoteProtocolInfo, PeerConnectionManager, PeerConnectionID, Direction, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the PrepareForConnection action.

            :returns: "ConnectionID", "AVTransportID", "RcsID"
        """
        arguments = {
            "RemoteProtocolInfo": RemoteProtocolInfo,
            "PeerConnectionManager": PeerConnectionManager,
            "PeerConnectionID": PeerConnectionID,
            "Direction": Direction,
        }

        out_params = self.call_action("PrepareForConnection", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("ConnectionID", "AVTransportID", "RcsID",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args




__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

import os

import click

from akit.environment.variables import AKIT_VARIABLES

@click.group("upnp", help="Group of UPNP utility commands.")
def group_akit_network_upnp():
    return

@click.command("query", help="Performs a UPNP msearch query directly against a device.")
@click.argument("ipaddress")
def command_akit_network_upnp_query(ipaddress):

    from akit.interop.upnp.upnpprotocol import msearch_query_host

    show_progress = False
    if AKIT_VARIABLES.AKIT_INTERACTIVE_CONSOLE:
        show_progress = True

    print("")

    found_device_info = msearch_query_host(ipaddress)
    
    print(found_device_info)

    return

@click.command("scan", help="Performs a UPNP msearch scan of the network.")
def command_akit_network_upnp_scan():

    from akit.interop.upnp.upnpprotocol import msearch_scan

    device_hints = []
    found_devices = {}
    matching_devices = {}

    show_progress = False
    if AKIT_VARIABLES.AKIT_INTERACTIVE_CONSOLE:
        show_progress = True

    print("")

    iter_found_devices, iter_matching_devices = msearch_scan(device_hints, show_progress=show_progress)
    found_devices.update(iter_found_devices)
    matching_devices.update(iter_matching_devices)

    print("")

    for fdusn, fdinfo in found_devices.items():
        dev_lines = [
            "DEVICE - {}".format(fdusn)
        ]
        for fdi_key, fdi_val in fdinfo.items():
            dev_lines.append("    {}: {}".format(fdi_key, fdi_val))
        dev_lines.append("")
        dev_info_out = os.linesep.join(dev_lines)
        print(dev_info_out)

    return

group_akit_network_upnp.add_command(command_akit_network_upnp_query)
group_akit_network_upnp.add_command(command_akit_network_upnp_scan)



"""

    NOTE: This is a code generated file.  This file should not be edited directly.
"""



from akit.aspects import AspectsUPnP, DEFAULT_UPNP_ASPECTS

from akit.extensible import LoadableExtension
from akit.interop.upnp.services.upnpserviceproxy import UpnpServiceProxy

class GroupRenderingControl1ServiceProxy(UpnpServiceProxy, LoadableExtension):
    """
        This is a code generated proxy class to the 'GroupRenderingControl1' service.
    """

    SERVICE_MANUFACTURER = 'SonosInc'
    SERVICE_TYPE = 'urn:schemas-upnp-org:service:GroupRenderingControl:1'

    SERVICE_DEFAULT_VARIABLES = {}

    SERVICE_EVENT_VARIABLES = {
        "GroupMute": { "data_type": "boolean", "default": None, "allowed_list": None},
        "GroupVolume": { "data_type": "ui2", "default": None, "allowed_list": None},
        "GroupVolumeChangeable": { "data_type": "boolean", "default": None, "allowed_list": None},
    }

    def action_GetGroupMute(self, InstanceID, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetGroupMute action.

            :returns: "CurrentMute"
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        out_params = self.call_action("GetGroupMute", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentMute",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetGroupVolume(self, InstanceID, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetGroupVolume action.

            :returns: "CurrentVolume"
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        out_params = self.call_action("GetGroupVolume", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentVolume",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_SetGroupMute(self, InstanceID, DesiredMute, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetGroupMute action.
        """
        arguments = {
            "InstanceID": InstanceID,
            "DesiredMute": DesiredMute,
        }

        self.call_action("SetGroupMute", arguments=arguments, aspects=aspects)

        return

    def action_SetGroupVolume(self, InstanceID, DesiredVolume, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetGroupVolume action.
        """
        arguments = {
            "InstanceID": InstanceID,
            "DesiredVolume": DesiredVolume,
        }

        self.call_action("SetGroupVolume", arguments=arguments, aspects=aspects)

        return

    def action_SetRelativeGroupVolume(self, InstanceID, Adjustment, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetRelativeGroupVolume action.

            :returns: "NewVolume"
        """
        arguments = {
            "InstanceID": InstanceID,
            "Adjustment": Adjustment,
        }

        out_params = self.call_action("SetRelativeGroupVolume", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewVolume",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_SnapshotGroupVolume(self, InstanceID, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SnapshotGroupVolume action.
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        self.call_action("SnapshotGroupVolume", arguments=arguments, aspects=aspects)

        return



"""
.. module:: upnpdevice
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Module containing the :class:`UpnpDevice` class and associated diagnostic.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>

"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import Optional, Union, TYPE_CHECKING

import threading
import weakref

from xml.etree.ElementTree import fromstring as xml_fromstring
from xml.etree.ElementTree import Element

import requests

from akit.exceptions import AKitNotOverloadedError, AKitServiceUnAvailableError
from akit.extensible import generate_extension_key
from akit.paths import normalize_name_for_path

from akit.interop.upnp.services.upnpserviceproxy import UpnpServiceProxy
from akit.interop.upnp.xml.upnpdevice1 import UpnpDevice1Device

# Types imported only for type checking purposes
if TYPE_CHECKING:
    from akit.interop.upnp.upnpfactory import UpnpFactory

UPNP_SERVICE1_NAMESPACE = "urn:schemas-upnp-org:service-1-0"

class UpnpDevice:
    """
        The UPNP Root device is the base device for the hierarchy that is
        associated with a unique network devices location.  The :class:`RootDevice`
        and its subdevices are linked by thier location url.

        http://www.upnp.org/specs/arch/UPnP-arch-DeviceArchitecture-v1.0.pdf
    """

    MANUFACTURER = "unknown"
    MODEL_NUMBER = "unknown"
    MODEL_DESCRIPTION = "unknown"

    def __init__(self):
        """
            Creates a root device object.
        """
        super(UpnpDevice, self).__init__()

        self._description = None
        self._host = None
        self._urlBase = None

        self._device_lock = threading.RLock()

        self._services_descriptions = {}
        self._services = {}

        return

    @property
    def description(self):
        """
            Returns the device description from the UPNP device.
        """
        return self._description

    @property
    def host(self):
        """
            Returns the host associated with the UPNP device.
        """
        return self._host

    @property
    def services(self):
        """
            Returns a list of service proxies associated with the UPNP device.
        """
        service_list = None

        self._device_lock.acquire()
        try:
            service_list = [svc for svc in self._services.values()]
        finally:
            self._device_lock.release()

        return service_list

    @property
    def services_descriptions(self):
        """
            Returns a table of service descriptions associated with the UPNP device.
        """
        return self._services_descriptions

    @property
    def URLBase(self):
        """
            Returns the base URL for the UPNP device that is used with any relative URLs in devices or service
            description documents.
        """
        return self._urlBase

    def get_service_description(self, service_type) -> Union[dict, None]:
        """
            Gets the service description for the specified service type.

            :param service_type: The service type to lookup the service description for.

            :returns: Returns the description information for the service type if found or None if not found.
        """
        svc_content = None

        devDesc = self.description

        for nxtsvc in devDesc.serviceList:
            if nxtsvc.serviceType == service_type:
                fullurl = self.URLBase.rstrip("/") + "/" + nxtsvc.SCPDURL.lstrip("/")

                resp = requests.get(fullurl)
                if resp.status_code == 200:
                    svc_content = resp.text
                    break

        return svc_content

    def lookup_service(self, serviceManufacturer: str, serviceType: str, allow_none: bool=False, factory: Optional["UpnpFactory"] = None) -> Union[UpnpServiceProxy, None]:
        """
            Looks up a service proxy based on the service manufacturer and service type specified.

            :param serviceManufacturer: The manufacturer associated with the device and service manufacturer.
            :param serviceType: The service type of the service to lookup.
            :param allow_none: If True, this API will not thrown an exception if the service cannot be found.

            :returns: The service proxy associated with the manufacturer and service type provided or None.
        """
        svc = None

        serviceManufacturer = normalize_name_for_path(serviceManufacturer)
        svckey = generate_extension_key(serviceManufacturer, serviceType)

        self._device_lock.acquire()
        try:
            if svckey in self._services:
                svc = self._services[svckey]
            elif factory is not None:
                self._device_lock.release()
                try:
                    svc = factory.create_service_instance(serviceManufacturer, serviceType)
                finally:
                    self._device_lock.acquire()

                if svc is not None:
                    self._services[svckey] = svc
        finally:
            self._device_lock.release()

        if svc is None and not allow_none:
            errmsg = "The service mfg={} serviceType={} was not found or is not available.".format(serviceManufacturer, serviceType)
            raise AKitServiceUnAvailableError(errmsg)

        return svc

    def to_dict(self, brief=False):
        """
            Creates a dictionary description of the device and its sub devices and services.
        """
        dval = None

        self._device_lock.acquire()
        try:
            desc = self._description

            if desc is not None:
                dval = desc.to_dict(brief=brief)

                dval["URLBase"] = self.URLBase

                if not brief:
                    serviceDescList = []
                    serviceList = dval["serviceList"]
                    for svc_info in serviceList:
                        svc_type = svc_info["serviceType"]
                        sdurl = self._urlBase.rstrip("/") + "/" + svc_info["SCPDURL"].lstrip("/")
                        sdesc = self._locked_process_full_service_description(sdurl)
                        sdesc["serviceType"] = svc_type
                        serviceDescList.append(sdesc)

                    dval["serviceDescriptionList"] = serviceDescList
        finally:
            self._device_lock.release()

        return dval

    def to_json(self, brief=False):
        """
            Creates a json description of the device and its sub devices and services.
        """
        json_str = None

        self._device_lock.acquire()
        try:
            desc = self._description

            if desc is not None:
                json_str = self._description.to_json(brief=brief)
        finally:
            self._device_lock.release()

        return json_str

    def _locked_populate_embedded_device_descriptions(self, factory: "UpnpFactory", description: UpnpDevice1Device):
        """
            Takes in a device description and processes it for embedded devices.

            :param factory: The UpnpFactory that has registered exentsion types that is used to instantiate
                            embedded devices when the UPNP device description is processed.
            :param description: The UpnpDevice1Device object that contains information about the device and its embedded devices.
        """
        # pylint: disable=no-self-use,unused-argument
        raise AKitNotOverloadedError("UpnpDevice._populate_embedded_devices: must be overridden.") from None

    def _locked_populate_services_descriptions(self, factory: "UpnpFactory", description: UpnpDevice1Device):
        """
            Takes in a device description and processes it for service descriptions.

            :param factory: The UpnpFactory that has registered exentsion types that is used to instantiate
                            services when the UPNP device description is processed.
            :param description: The UpnpDevice1Device object that contains information about the device and its services.
        """
        # pylint: disable=protected-access
        for serviceInfo in description.serviceList:
            serviceManufacturer = normalize_name_for_path(serviceInfo.serviceManufacturer)
            serviceType = serviceInfo.serviceType

            svckey = generate_extension_key(serviceManufacturer, serviceType)
            if serviceType not in self._services_descriptions:
                self._services_descriptions[svckey] = serviceInfo

                svc_inst = factory.create_service_instance(serviceManufacturer, serviceType)
                if svc_inst is not None:
                    device_ref = weakref.ref(self)
                    svc_inst._proxy_link_service_to_device(device_ref, serviceInfo)
                    self._services[svckey] = svc_inst
            else:
                svc_inst = self._services[svckey]

        return

    def _locked_process_full_service_description(self, sdurl: str):
        """
            Downloads and processes the service description document.

            :param sdurl: The url of the service description document to process.
        """
        svcdesc = None

        resp = requests.get(sdurl)
        if resp.status_code == 200:
            svcdesc = {}

            namespaces = {"": UPNP_SERVICE1_NAMESPACE}

            try:
                xml_content = resp.text
                descDoc = xml_fromstring(xml_content)

                specVersionNode = descDoc.find("specVersion", namespaces=namespaces)
                verInfo = self._locked_process_node_spec_version(specVersionNode, namespaces=namespaces)
                svcdesc["specVersion"] = verInfo

                serviceStateTableNode = descDoc.find("serviceStateTable", namespaces=namespaces)
                variablesTable, typesTable, eventsTable = self._locked_process_node_state_table(serviceStateTableNode, namespaces=namespaces)
                svcdesc["variablesTable"] = variablesTable
                svcdesc["typesTable"] = typesTable
                svcdesc["eventsTable"] = eventsTable

                actionListNode = descDoc.find("actionList", namespaces=namespaces)
                actionsTable = self._locked_process_node_action_list(actionListNode, namespaces=namespaces)
                svcdesc["actionsTable"] = actionsTable
            except:
                print("Service Description Failure: %s" % sdurl)
                raise

        return svcdesc

    def _locked_process_node_action_list(self, actionListNode: Element, namespaces: Optional[dict] = None) -> dict:
        """
            Processes the node action list and creates a dictionary of the actions supported by the service.

            :param actionListNode: The parent node for all the action list node.
            :param namespaces: A dictionary of namespaces to use when processing the XML.
        """
        # pylint: disable=no-self-use
        actionTable = {}

        actionNodeList = actionListNode.findall("action", namespaces=namespaces)
        for actionNode in actionNodeList:
            actionName = actionNode.find("name", namespaces=namespaces).text
            actionArgs = {}

            argumentListNode = actionNode.find("argumentList", namespaces=namespaces)
            if argumentListNode is not None:
                argumentNodeList = argumentListNode.findall("argument", namespaces=namespaces)
                for argumentNode in argumentNodeList:
                    argName = argumentNode.find("name", namespaces=namespaces).text
                    argDirection = argumentNode.find("direction", namespaces=namespaces).text

                    argRelStateVar = None
                    argRelStateNode = argumentNode.find("relatedStateVariable", namespaces=namespaces)
                    if argRelStateNode is not None:
                        argRelStateVar = argRelStateNode.text

                    actionArgs[argName] = {
                        "name": argName,
                        "direction": argDirection,
                        "relatedStateVariable": argRelStateVar
                        }

            actionTable[actionName] = {
                    "name": actionName,
                    "arguments" : actionArgs
                }

        return actionTable

    def _locked_process_node_spec_version(self, specVersionNode: Element, namespaces: Optional[dict] = None) -> dict:
        # pylint: disable=no-self-use
        verInfo = {}

        verInfo["major"] = specVersionNode.find("major", namespaces=namespaces).text
        verInfo["minor"] = specVersionNode.find("minor", namespaces=namespaces).text

        return verInfo

    def _locked_process_node_state_table(self, serviceStateTableNode: Element, namespaces: Optional[dict] = None) -> dict:
        # pylint: disable=no-self-use
        variablesTable = {}
        typesTable = {}
        eventsTable = {}

        stateVariableList = serviceStateTableNode.findall("stateVariable", namespaces=namespaces)
        for stateVariableNode in stateVariableList:
            name = stateVariableNode.find("name", namespaces=namespaces).text
            dataType = stateVariableNode.find("dataType", namespaces=namespaces).text

            varInfo = { "name": name, "dataType": dataType}
            if name.startswith("A_ARG_TYPE_"):
                typesTable[name] = varInfo
            else:
                variablesTable[name] = varInfo

                sendEvents = "no"
                if "sendEvents" in stateVariableNode.attrib:
                    sendEvents = stateVariableNode.attrib["sendEvents"]

                if sendEvents == "yes":
                    eventsTable[name] = varInfo

        return variablesTable, typesTable, eventsTable
